######################################################################################
# Copyright (c) 2023 Orange - All Rights Reserved                             #
# * This software is the confidential and proprietary information of Orange.         #
# * You shall not disclose such Restricted Information and shall use it only in      #
#   accordance with the terms of the license agreement you entered into with Orange  #
#   named the "kuplift - Python Library Evaluation License".                          #
# * Unauthorized copying of this file, via any medium is strictly prohibited.        #
# * See the "LICENSE.md" file for more details.                                      #
######################################################################################
from kuplift import FeatureSelection


def test_filter(test_dataframe):
    fs = FeatureSelection()

    features = list(test_dataframe.columns[:-2])
    important_vars = fs.filter(test_dataframe[features], test_dataframe["segment"], test_dataframe["visit"])
    assert important_vars == {
        "x2_informative": 0,
        "x4_informative": 0,
        "x6_informative": 0,
        "x7_informative": 0,
        "x8_informative": 0,
        "x11_irrelevant": 0,
        "x12_irrelevant": 0,
        "x13_irrelevant": 0,
        "x14_irrelevant": 0,
        "x15_irrelevant": 0,
        "x16_irrelevant": 0,
        "x17_irrelevant": 0,
        "x18_irrelevant": 0,
        "x19_irrelevant": 0,
        "x20_irrelevant": 0,
        "x21_irrelevant": 0,
        "x22_irrelevant": 0,
        "x23_irrelevant": 0,
        "x24_irrelevant": 0,
        "x25_irrelevant": 0,
        "x26_irrelevant": 0,
        "x27_irrelevant": 0,
        "x28_irrelevant": 0,
        "x29_irrelevant": 0,
        "x30_irrelevant": 0,
        "x1_informative": 0.010107186288141222,
        "x3_informative": 0.010230804170840865,
        "x5_informative": 0.010579495308748063,
        "x10_informative": 0.010998393563985725,
        "x9_informative": 0.011337824503076499,
        "x35_uplift_increase": 0.014427997839683255,
        "x34_uplift_increase": 0.015052299093348126,
        "x33_uplift_increase": 0.015064719742149577,
        "x32_uplift_increase": 0.016302396142469408,
        "x31_uplift_increase": 0.017155249895371802,
        "x36_uplift_increase": 0.018097298951444143,
    }



######################################################################################
# Copyright (c) 2023 Orange - All Rights Reserved                             #
# * This software is the confidential and proprietary information of Orange.         #
# * You shall not disclose such Restricted Information and shall use it only in      #
#   accordance with the terms of the license agreement you entered into with Orange  #
#   named the "kuplift - Python Library Evaluation License".                          #
# * Unauthorized copying of this file, via any medium is strictly prohibited.        #
# * See the "LICENSE.md" file for more details.                                      #
######################################################################################
from math import log
import numpy as np
import random
from .helperfunctions import (
    log_fact,
    universal_code_natural_numbers,
)
from .tree import _Tree


class _UpliftTreeClassifier(_Tree):
    """Private child class

    Parameters
    ----------
    data : pd.Dataframe
        Dataframe containing feature variables.
    treatment_col : pd.Series
        Treatment column.
    y_col : pd.Series
        Outcome column.
    control_name: int or str
        The name of the control value in the treatment column
    """

    def __init__(self,control_name=None):
        super().__init__()

    def grow_tree(self,data, treatment_col, y_col):
        super().__initializeVars__(data.copy(), treatment_col.copy(), y_col.copy())
        # In case if we have a new attribute for splitting
        prob_kt_plus_one = (
            universal_code_natural_numbers(self.k_t + 1)
            - log_fact(self.k_t + 1)
            + (self.k_t + 1) * log(self.k)
        )
        prob_of_attribute_selection_among_subset_attributes_plus_one = log(
            self.k_t + 1
        ) * (len(self.internal_nodes) + 1)

        encoding_of_being_an_internal_node_plus_one = (
            self.encoding_of_being_an_internal_node + log(2)
        )

        # When splitting a node to 2 nodes, the number of leaf nodes is
        # incremented only by one, since the parent node was leaf
        # and is now internal.
        # 2 for two extra leaf nodes multiplied by 2 for w. Total = 4.
        encoding_of_being_a_leaf_node_and_containing_te_plus_two = (
            self.encoding_of_being_a_leaf_node_and_containing_te + (2 * log(2))
        )

        encoding_of_internal_and_leaves_and_w_with_extra_nodes = (
            encoding_of_being_an_internal_node_plus_one
            + encoding_of_being_a_leaf_node_and_containing_te_plus_two
        )

        i = 0
        while True:
            node_vs_best_attribute_corresponding_to_the_best_cost = {}
            node_vs_best_cost = {}
            # Dictionary containing Nodes as key and their values are another
            # dictionary each with attribute:CostSplit
            node_vs_candidate_splits_costs = {}
            for terminal_node in self.terminal_nodes:
                # This if condition is here to not to repeat calculations of
                # candidate splits
                if terminal_node.candidate_splits_vs_criterion is None:
                    node_vs_candidate_splits_costs[
                        terminal_node
                    ] = (
                        terminal_node.discretize_vars_and_get_attributes_splits_costs()
                    )
                else:
                    node_vs_candidate_splits_costs[
                        terminal_node
                    ] = terminal_node.candidate_splits_vs_criterion.copy()

                if len(node_vs_candidate_splits_costs[terminal_node]) == 0:
                    continue

                # Update Costs
                list_of_attribute_splits_improving_tree_criterion = []
                for attribute in node_vs_candidate_splits_costs[terminal_node]:
                    if attribute in self.feature_subset:
                        node_vs_candidate_splits_costs[terminal_node][
                            attribute
                        ] += (
                            self.prob_kt
                            + self.prob_attribute_selection
                            + encoding_of_internal_and_leaves_and_w_with_extra_nodes
                            + self.leaf_prior
                            + self.tree_likelihood
                            + self.prior_of_internal_nodes
                        )
                    else:
                        node_vs_candidate_splits_costs[terminal_node][
                            attribute
                        ] += (
                            prob_kt_plus_one
                            + encoding_of_internal_and_leaves_and_w_with_extra_nodes
                            + prob_of_attribute_selection_among_subset_attributes_plus_one
                            + self.leaf_prior
                            + self.tree_likelihood
                            + self.prior_of_internal_nodes
                        )

                    if (
                        node_vs_candidate_splits_costs[terminal_node][
                            attribute
                        ]
                        < self.tree_criterion
                    ):
                        list_of_attribute_splits_improving_tree_criterion.append(
                            attribute
                        )
                if len(list_of_attribute_splits_improving_tree_criterion) == 0:
                    continue
                key_of_the_minimal_val = random.choice(
                    list_of_attribute_splits_improving_tree_criterion
                )  # key_of_the_minimal_val is the attribute name

                node_vs_best_attribute_corresponding_to_the_best_cost[
                    terminal_node
                ] = key_of_the_minimal_val
                node_vs_best_cost[
                    terminal_node
                ] = node_vs_candidate_splits_costs[terminal_node][
                    key_of_the_minimal_val
                ]

            if len(list(node_vs_best_cost)) == 0:
                break
            optimal_node_attribute_to_split_up = random.choice(
                list(node_vs_best_cost)
            )
            optimal_val = node_vs_best_cost[optimal_node_attribute_to_split_up]
            optimal_node = optimal_node_attribute_to_split_up
            optimal_attribute = (
                node_vs_best_attribute_corresponding_to_the_best_cost[
                    optimal_node_attribute_to_split_up
                ]
            )

            if optimal_val < self.tree_criterion:
                self.tree_criterion = optimal_val
                if optimal_attribute not in self.feature_subset:
                    self.feature_subset.append(optimal_attribute)
                    self.k_t += 1
                new_left_leaf, new_right_leaf = optimal_node.perform_split(
                    optimal_attribute
                )
                self.terminal_nodes.append(new_left_leaf)
                self.terminal_nodes.append(new_right_leaf)
                self.internal_nodes.append(optimal_node)
                self.terminal_nodes.remove(optimal_node)

                self.calc_criterion()
            else:
                break

        self.tree_criterion = (
            self.prob_kt
            + self.encoding_of_being_an_internal_node
            + self.prob_attribute_selection
            + self.prior_of_internal_nodes
            + self.encoding_of_being_a_leaf_node_and_containing_te
            + self.leaf_prior
            + self.tree_likelihood
        )


class BayesianRandomForest:
    """
    The BayesianRandomForest class implements the UB-RF algorithm described in:
    Rafla, M., Voisine, N., Crémilleux, B., & Boullé, M. (2023, May).
    A Non-Parametric Bayesian Decision Trees for Uplift modelling. In PAKDD.

    Parameters
    ----------
    data : pd.Dataframe
        Dataframe containing data.
    treatment_col : pd.Series
        Treatment column.
    outcome_col : pd.Series
        Outcome column.
    n_trees : int, default 10
        Number of trees in a forest.
    vars_subset : bool, default False
        Use a random subset of the variables for each tree in the forest.
    random_state : int, default 10
        Seed used by the random number generator.
    """

    def __init__(
        self,
        n_trees=10,
        vars_subset=False,
        random_state=10,
    ):
        self.list_of_trees = []
        
        self.n_trees=n_trees

        self.vars_subset=vars_subset

        self.random_state=random_state
        random.seed(self.random_state)

        self.treatment_name='treatment'

        self.outcome_name='outcome'

    def fit(self,data,treatment_col,y_col):
        """Fit a decision tree algorithm."""
        # data.loc[:,self.treatment_name]=treatment_col
        # data.loc[:,self.outcome_name]=y_col
        # self.data = data
        
        if self.vars_subset: # Randomly select columns for the data
            cols = list(self.data.columns)
            # cols.remove(self.treatment_name)
            # cols.remove(self.outcome_name)
            cols = random.sample(cols, int(np.sqrt(len(cols))))
            # data = data[cols + [self.treatment_name, self.outcome_name]]
            data = data[cols]

        for i in range(self.n_trees):
            Tree = _UpliftTreeClassifier()
            self.list_of_trees.append(Tree)

        for tree in self.list_of_trees:
            tree.grow_tree(data.copy(),treatment_col.copy(), y_col.copy())

    def predict(self, X_test, weighted_average=False):
        """
        Predict the uplift value for each example in X_test.

        Parameters
        ----------
        X_test : pd.Dataframe
            Dataframe containing test data.
        weighted_average : bool, default False
            Give a weight for the predictions of each tree according to its cost.

        Returns
        -------
        y_pred_list : (ndarray, shape=(num_samples, 1))
            An array containing the predicted uplift for each sample.
        """
        if not weighted_average:
            list_of_preds = []

            for tree in self.list_of_trees:
                list_of_preds.append(np.array(tree.predict(X_test)))
            return np.mean(list_of_preds, axis=0)
        else:
            list_of_criterion = []
            list_of_preds = []
            for tree in self.list_of_trees:
                list_of_criterion.append(np.array(tree.tree_criterion))
                list_of_preds.append(np.array(tree.predict(X_test)))

            sum_of_criterions = sum(list_of_criterion)
            for i in range(len(list_of_criterion)):
                list_of_criterion[i] = sum_of_criterions / list_of_criterion[i]
            sum_of_weights = sum(list_of_criterion)

            list_of_weights = []
            for i in range(len(list_of_criterion)):
                list_of_weights.append(list_of_criterion[i] / sum_of_weights)
            return np.average(list_of_preds, axis=0, weights=list_of_weights)



######################################################################################
# Copyright (c) 2023 Orange - All Rights Reserved                             #
# * This software is the confidential and proprietary information of Orange.         #
# * You shall not disclose such Restricted Information and shall use it only in      #
#   accordance with the terms of the license agreement you entered into with Orange  #
#   named the "kuplift - Python Library Evaluation License".                          #
# * Unauthorized copying of this file, via any medium is strictly prohibited.        #
# * See the "LICENSE.md" file for more details.                                      #
######################################################################################
import pandas as pd
import multiprocessing as mp
from .helperfunctions import preprocess_data
from .umodl_search_algorithm import execute_greedy_search_and_post_opt


class UnivariateEncoding:
    """
    The UnivariateEncoding class implements the UMODL algorithm for uplift data encoding described in:
    Rafla, M., Voisine, N., Crémilleux, B., & Boullé, M.
    (2023, March). A non-parametric bayesian approach for uplift
    discretization and feature selection. ECML PKDD
    """

    def __init__(self,control_name=None):
        self.var_vs_disc = {}
        self.treatment_name = "treatment"
        self.outcome_name = "outcome"
        self.column_names=[]
        self.control_name=control_name
        self.disc_details={}
    
    def get_features_importance_details(self):
        # Prepare lists to capture multi-index levels
        outer_keys = []
        inner_keys = []
        values = []

        # Loop through the dictionary to populate the lists
        for key, sublist in self.disc_details.items():
            for idx, val in enumerate(sublist):
                outer_keys.append(key)
                inner_keys.append(idx)
                values.append(val)

        # Create the multi-index
        multi_index = pd.MultiIndex.from_tuples(list(zip(outer_keys, inner_keys)), names=["key", ""])

        # Convert the lists to DataFrame
        df = pd.DataFrame(values, columns=["interval", "P(Y|T=1)", "P(Y|T=0)", "uplift"], index=multi_index)
        return df
    
    def fit_transform(
        self, data, treatment_col, y_col, parallelized=False, num_processes=5
    ):
        """
        fit_transform() learns a discretisation model using UMODL and
        transforms the data.

        Parameters
        ----------
        data : pd.Dataframe
            Dataframe containing feature variables.
        treatment_col : pd.Series
            Treatment column.
        y_col : pd.Series
            Outcome column.
        parallelized : bool, default False
            Whether to run the code on several processes.
        num_processes : int, default 5
            Number of processes to use in parallel.

        Returns
        -------
        pd.Dataframe
            Pandas Dataframe that contains encoded data.
        """
        self.fit(data, treatment_col, y_col, parallelized, num_processes)
        data = self.transform(data)
        return data
    
    def fit(
        self, data, treatment_col, y_col, parallelized=False, num_processes=5
    ):
        """
        fit() learns a discretisation model using the UMODL approach.

        Parameters
        ----------
        data : pd.Dataframe
            Dataframe containing feature variables.
        treatment_col : pd.Series
            Treatment column.
        y_col : pd.Series
            Outcome column.
        parallelized : bool, default False
            Whether to run the code on several processes.
        num_processes : int, default 5
            Number of processes to use in parallel.
        """
        self.column_names=data.columns
        if treatment_col.name in self.column_names:
            raise Exception("The treatment column is in the data, it should be passed in 'treatment_col' argument")
        if y_col.name in self.column_names:
            raise Exception("The outcome column is in the data, it should be passed in 'y_col' argument")
        
        
        data = data.assign(**{self.treatment_name: treatment_col.copy()})
        data = data.assign(**{self.outcome_name: y_col.copy()})
        
        #dealing with control name
        if self.control_name != None:
            trt_vals=list(data[self.treatment_name].unique())
            
            #Verify that control name is in the treatment column, else raise an exception
            if self.control_name not in trt_vals:
                raise Exception("the control name is not in the treatment column")
            data[self.treatment_name] = data[self.treatment_name].replace(self.control_name,0)

            trt_vals.remove(self.control_name)
            #the other value will be considered as the treatment
            data[self.treatment_name] = data[self.treatment_name].replace(trt_vals[0],1)
        

        
        cols = list(data.columns)
        if self.treatment_name in cols:
            cols.remove(self.treatment_name)
        if self.outcome_name in cols:    
            cols.remove(self.outcome_name)

        data = data[cols + [self.treatment_name, self.outcome_name]]
        data = preprocess_data(data, self.treatment_name, self.outcome_name)
        var_vs_importance = {}
        self.var_vs_disc = {}

        if parallelized == True:
            pool = mp.Pool(processes=num_processes)

            arguments_to_pass_in_parallel = []
            for col in cols:
                arguments_to_pass_in_parallel.append(
                    [data[[col, self.treatment_name, self.outcome_name]],True]
                )
            list_of_tuples_feature_vs_importance = pool.starmap(
                execute_greedy_search_and_post_opt,
                arguments_to_pass_in_parallel,
            )
            pool.close()
            print(list_of_tuples_feature_vs_importance)
            for el in list_of_tuples_feature_vs_importance:
                col = el[2]
                if len(el[1]) == 1:
                    self.var_vs_disc[col] = None
                else:
                    self.var_vs_disc[col] = el[1][:-1]

        else:
            for col in cols:
                (
                    var_vs_importance[col],
                    self.var_vs_disc[col],
                    col_name,
                    self.disc_details
                ) = execute_greedy_search_and_post_opt(
                    data[[col, self.treatment_name, self.outcome_name]]
                )
                if len(self.var_vs_disc[col]) == 1:
                    self.var_vs_disc[col] = None
                else:
                    self.var_vs_disc[col] = self.var_vs_disc[col][:-1]
    def transform(self, data):
        """
        transform() applies the discretisation model learned by the
        fit() method.

        Parameters
        ----------
        data : pd.Dataframe
            Dataframe containing feature variables.

        Returns
        -------
        pd.Dataframe
            Pandas Dataframe that contains encoded data.
        """
        
        cols = list(data.columns)
        
        not_in_self_column_names = any(x not in self.column_names for x in cols)
        if not_in_self_column_names==True:
            raise Exception("The data to be transformed contains unknown columns")

        if self.treatment_name in cols:
            cols.remove(self.treatment_name)
        if self.outcome_name in cols:
            cols.remove(self.outcome_name)        
        for col in cols:
            if self.var_vs_disc[col] is None:
                data.drop(col, inplace=True, axis=1)
            else:
                minBoundary = min(
                    data[col].min(), self.var_vs_disc[col][0] - 0.001
                )
                maxBoundary = max(
                    data[col].max(), self.var_vs_disc[col][-1] + 0.001
                )
                
                data_copy = data.copy()
                data_copy[col] = pd.cut(
                    data_copy[col],
                    bins=[minBoundary] + self.var_vs_disc[col] + [maxBoundary],
                )
                data = data_copy.copy()

                data[col] = data[col].astype("category")
                data[col] = data[col].cat.codes
        return data



######################################################################################
# Copyright (c) 2023 Orange - All Rights Reserved                             #
# * This software is the confidential and proprietary information of Orange.         #
# * You shall not disclose such Restricted Information and shall use it only in      #
#   accordance with the terms of the license agreement you entered into with Orange  #
#   named the "kuplift - Python Library Evaluation License".                          #
# * Unauthorized copying of this file, via any medium is strictly prohibited.        #
# * See the "LICENSE.md" file for more details.                                      #
######################################################################################
from math import log

_log_fact_table = []


def log_fact(n):
    """Compute log(fact(n)).

    Parameters
    ----------
    n : int
        Number on which to do the calculation.

    Returns
    -------
    float
        Value of log(fact(n)).
    """
    # Use approximation for large n
    if n > 1e6:
        raise ValueError("Value of n is too large")
    # computation of values, tabulation in private array
    else:
        s = len(_log_fact_table)
        if n >= s:
            if s == 0:
                _log_fact_table.append(0)
            size = len(_log_fact_table)
            while size <= n:
                _log_fact_table.append(log(size) + _log_fact_table[size - 1])
                size = size + 1
        return _log_fact_table[n]


def log_2_star(k):
    """Computes the term log_2*(k)=log_2(k) + log_2(log_2(k)) + ...
    of Rissanen's code for integers so long as the terms are positive.

    Parameters
    ----------
    k : int
        Number on which to do the calculation.

    Returns
    -------
    float
        Value of the result.
    """
    d_log2 = log(2.0)
    d_cost = 0.0
    d_logI = log(1.0 * k) / d_log2

    if k < 1:
        raise ValueError(
            "Universal code is defined for natural numbers over 1"
        )
    else:
        while d_logI > 0:
            d_cost += d_logI
            d_logI = log(d_logI) / d_log2
        return d_cost


def log_binomial_coefficient(n, k):
    """Computes the log of the binomial coefficient (n
                                                     k)
    (log of the total number of combinations of k elements from n).

    Parameters
    ----------
    n : int
        Total number of elements.
    k : int
        Number of selected elements.

    Returns
    -------
    float
        Value of the result.
    """
    # start_counter(5)
    nf = log_fact(n)
    kf = log_fact(k)
    nkf = log_fact(n - k)
    # stop_counter(5)
    return (nf - nkf) - kf


def universal_code_natural_numbers(k):
    """Compute the universal code for integers presented by Rissanen in
    'A Universal Prior for Integers and Estimation by Minimum Description
    Length', Rissanen 1983.

    Parameters
    ----------
    k : int

    Returns
    -------
    float
        Value of the result.
    """
    dC0 = 2.86511  # First value computed following the given estimation
    # formula, as e(3)=65536 + d_log2^5 / (1-d_log2)
    d_log2 = log(2.0)

    if k < 1:
        raise ValueError(
            "Universal code is defined for natural numbers over 1"
        )
    else:
        d_cost = log(dC0) / d_log2  # Initialize code length cost to log_2(dC0)
        d_cost += log_2_star(k)  # Add log_2*(k)
        d_cost *= d_log2  # Go back to the natural log
        return d_cost


def preprocess_data(data, treatment_col="segment", y_col="visit"):
    """preprocess data

    Parameters
    ----------
    data : pd.Dataframe
        Dataframe containing feature variables.
    treatment_col : pd.Series, default "segment"
        Treatment column.
    y_col : pd.Series, default "visit"
        Outcome column.

    Returns
    -------
    pd.Dataframe
        Pandas Dataframe that contains encoded data.
    """
    cols = data.columns
    num_cols = list(data._get_numeric_data().columns)

    if treatment_col in num_cols:
        num_cols.remove(treatment_col)
    if y_col in num_cols:
        num_cols.remove(y_col)

    num_col_index = 0
    while num_col_index < len(num_cols):
        num_col = num_cols[num_col_index]
        if len(data[num_col].value_counts()) < 1000:
            num_cols.remove(num_col)
        else:
            data[num_col] = data[num_col].fillna(data[num_col].min() - 1)
            num_col_index = num_col_index + 1

    categorical_cols = list(set(cols) - set(num_cols))
    if treatment_col in categorical_cols:
        categorical_cols.remove(treatment_col)
    if y_col in categorical_cols:
        categorical_cols.remove(y_col)
    for cat_col in categorical_cols:
        data[cat_col] = data[cat_col].fillna("NAN_VAL")
        dict_val_vs_uplift = {}
        for val in data[cat_col].value_counts().index:
            if val == "NAN_VAL":
                continue
            dataset_slice = data[data[cat_col] == val]
            t0j0 = dataset_slice[
                (dataset_slice[treatment_col] == 0)
                & (dataset_slice[y_col] == 0)
            ].shape[0]
            t0j1 = dataset_slice[
                (dataset_slice[treatment_col] == 0)
                & (dataset_slice[y_col] == 1)
            ].shape[0]
            t1j0 = dataset_slice[
                (dataset_slice[treatment_col] == 1)
                & (dataset_slice[y_col] == 0)
            ].shape[0]
            t1j1 = dataset_slice[
                (dataset_slice[treatment_col] == 1)
                & (dataset_slice[y_col] == 1)
            ].shape[0]

            if (t1j1 + t1j0) == 0:
                uplift_in_this_slice = -1
            elif (t0j1 + t0j1) == 0:
                uplift_in_this_slice = 0
            else:
                uplift_in_this_slice = (t1j1 / (t1j1 + t1j0)) - (
                    t0j1 / (t0j1 + t0j1)
                )
            dict_val_vs_uplift[val] = uplift_in_this_slice
        ordered_dict = {
            k: v
            for k, v in sorted(
                dict_val_vs_uplift.items(), key=lambda item: item[1]
            )
        }

        data[cat_col] = data[cat_col].replace(["NAN_VAL"], -1)
        encoded_i = 0
        for k, v in ordered_dict.items():
            data[cat_col] = data[cat_col].replace([k], encoded_i)
            encoded_i += 1
    data[treatment_col] = data[treatment_col].astype(str)
    return data



######################################################################################
# Copyright (c) 2023 Orange - All Rights Reserved                             #
# * This software is the confidential and proprietary information of Orange.         #
# * You shall not disclose such Restricted Information and shall use it only in      #
#   accordance with the terms of the license agreement you entered into with Orange  #
#   named the "kuplift - Python Library Evaluation License".                          #
# * Unauthorized copying of this file, via any medium is strictly prohibited.        #
# * See the "LICENSE.md" file for more details.                                      #
######################################################################################
from kuplift.bayesian_decision_tree import BayesianDecisionTree
from kuplift.bayesian_random_forest import BayesianRandomForest
from kuplift.feature_selection import FeatureSelection
from kuplift.univariate_encoding import UnivariateEncoding
from kuplift.optimized_univariate_encoding import OptimizedUnivariateEncoding

__all__ = [
    "BayesianDecisionTree",
    "BayesianRandomForest",
    "FeatureSelection",
    "UnivariateEncoding",
    "OptimizedUnivariateEncoding"
]



######################################################################################
# Copyright (c) 2023 Orange - All Rights Reserved                             #
# * This software is the confidential and proprietary information of Orange.         #
# * You shall not disclose such Restricted Information and shall use it only in      #
#   accordance with the terms of the license agreement you entered into with Orange  #
#   named the "kuplift - Python Library Evaluation License".                          #
# * Unauthorized copying of this file, via any medium is strictly prohibited.        #
# * See the "LICENSE.md" file for more details.                                      #
######################################################################################

"""Optimized Univariate Encoding

This module contains everything needed to make univariate variable transformation
optimized through the use of the C++ implementation of 'umodl'. It calls the
'umodl' executable as a subprocess indirectly by the use the 'umodl' library.

The main class of this module is 'OptimizedUnivariateEncoding'.

An example code is in examples/optimized_univariate_encoding.py.
"""

import pathlib
import tempfile
from warnings import warn
import math
import json
from typing import Optional, Sequence, Union
from itertools import starmap
from dataclasses import dataclass
from abc import ABC, abstractmethod
import khiops.sklearn.dataset
import pandas as pd
from umodl import run_umodl


class Partition(ABC):
    @property
    @abstractmethod
    def parts(self):
        pass

    def __iter__(self):
        return iter(self.parts)
    

class ValGrp(list):
    def __str__(self):
        return "{%s}" % ", ".join(self)


class ValGrpPartition(Partition):
    """Partition of type 'value groups'.
    
    Attributes
    ----------
        groups: Sequence[ValGrp]
            The groups. Each group is an iterable of its values.
        defaultgroupindex: int
            The group index affected to transformed elements when they do not explicitly appear in any group.
    """

    def __init__(self, groups: Sequence[ValGrp], defaultgroupindex: int):
        if not groups:
            raise ValueError("there must be at least one group")
        if defaultgroupindex < 0 or defaultgroupindex >= len(groups):
            raise ValueError(f"default group index is {defaultgroupindex} but groups are numbered from 0 to {len(groups) - 1}")
        self.groups = groups
        self.defaultgroupindex = defaultgroupindex

    @property
    def parts(self):
        return self.groups

    def transform(self, col):
        return col.transform(self.transform_elem)

    def transform_elem(self, elem):
        for i, group in enumerate(self.groups):
            if str(elem) in group:
                return i
        return self.defaultgroupindex
    
    def __repr__(self):
        return f"ValGrpPartition({self.groups!r}, {self.defaultgroupindex!r})"

    def __str__(self):
        return """
Value group partition
    {ngroups} groups ("*" indicates the default group):
{groups}
"""[1:-1].format(
    ngroups=len(self.groups),
    groups="\n".join(f"      {'*' if i == self.defaultgroupindex else ' '} - {group}" for i, group in enumerate(self.groups))
)


@dataclass
class Interval:
    lower: Optional[float] = None
    upper: Optional[float] = None
        
    @property
    def catches_missing(self):
        return self.lower is None or self.upper is None
    
    def __contains__(self, x):
        return not self.catches_missing and (self.lower <= x < self.upper)
    
    def __str__(self):
        return "[]" if self.catches_missing else f"[{self.lower}, {self.upper}["
    
    def __bool__(self):
        return not self.catches_missing


class IntervalPartition(Partition):
    """Partition of type 'intervals'.
    
    Attributes
    ----------
        intervals: Sequence[Interval]
            The intervals. Each interval is a pair defining its lower bound and its upper bound (in that order).
            The exception to this rule is the empty interval representing 'MISSING' values. If present, it must be
            the first interval of the sequence.
    """

    def __init__(self, intervals: Sequence[Interval]):
        if not intervals:
            raise ValueError("there must be at least one interval")
        if intervals[0].catches_missing:
            if len(intervals) >= 1:
                intervals[1].lower = -math.inf
                intervals[-1].upper = math.inf
        else:
            intervals[0].lower = -math.inf
            intervals[-1].upper = math.inf
        self.intervals = intervals

    @property
    def parts(self):
        return self.intervals

    def transform(self, col):
        return col.transform(self.transform_elem)

    def transform_elem(self, elem):
        if not isinstance(elem, (int, float)) or math.isnan(elem):
            return 0
        for i, interval in enumerate(self.intervals):
            if elem in interval:
                return i
            
    def __repr__(self):
        return f"IntervalPartition({self.intervals!r})"
    
    def __str__(self):
        return """
Interval partition
    {nintervals} intervals:
{intervals}
"""[1:-1].format(
    nintervals=len(self.intervals),
    intervals="\n".join(f"      - {interval}" for interval in self.intervals)
)
    

@dataclass
class TargetTreatmentPair:
    """Target-treatment pair.

    Used to identify both a target and a treatment.
    This class only exists for the purpose of formatting.
    """

    target: object
    treatment: object

    def __hash__(self):
        return hash((self.target, self.treatment))
    
    def __str__(self):
        return f"P({self.target}|{self.treatment})"


class OptimizedUnivariateEncoding:
    """
    The OptimizedUnivariateEncoding class makes use of the external umodl tool hosted at https://github.com/UData-Orange/umodl.

    Attributes
    ----------
    model: dict mapping str to ValGrpPartition or IntervalPartition
        The model generated by the 'umodl' executable.
        It describes the partitioning of values of informative variables into groups or intervals.
        It maps the informative variable names to value partitions.

    levels: list of (str, float) pairs
        (variable-name, variable-level) pairs in decreasing level order.

    variable_cols: DataFrame
        The data columns of all variables.
        This means all the data from the dataset but the treatment and target columns.

    treatment_col: Series
        The treatment column from the dataset.

    target_col: Series
        The target column from the dataset.
    """

    def __init__(self):
        self.model: dict[str, Union[ValGrpPartition, IntervalPartition]] = {}
        self.levels: list[tuple[str, float]] = []
        self.variable_cols = None
        self.treatment_col = None
        self.target_col = None

    @property
    def input_variables(self):
        """list of str
        
        The names of the variables.
        """
        return self.variable_cols.columns.to_list()
    
    @property
    def treatment_name(self):
        """str
        
        The name of the treatment column.
        """
        return self.treatment_col.name
    
    @property
    def target_name(self):
        """str
        
        The name of the target column.
        """
        return self.target_col.name
    
    @property
    def treatment_modalities(self):
        """list
        
        All the different treatments from the dataset.
        """
        return list(self.treatment_col.unique())
    
    @property
    def target_modalities(self):
        """list
        
        All the different targets from the dataset.
        """
        return list(self.target_col.unique())
    
    @property
    def target_treatment_pairs(self):
        """list of TargetTreatmentPair
        
        All (target, treatment) pairs.
        """
        return [TargetTreatmentPair(target, treatment) for target in self.target_modalities for treatment in self.treatment_modalities]

    def fit_transform(self, data, treatment_col, target_col, maxpartnumber = None):
        """Learn a discretisation model using UMODL and transform the data.

        Parameters
        ----------
        data: pd.DataFrame
            Dataframe containing feature variables. Categorical
            variables should have the object dtype, otherwise they
            are processed as numerical variables.
        treatment_col: pd.Series
            Treatment column.
        target_col: pd.Series
            Outcome column.
        maxpartnumber: int, default=None
            The maximal number of intervals or groups. None means default to the 'umodl' program default.

        Returns
        -------
        pd.Dataframe
            Pandas Dataframe that contains encoded data.
        """
        self.fit(data, treatment_col, target_col, maxpartnumber)
        return self.transform(data)
    
    def fit(self, data, treatment_col, target_col, maxpartnumber = None):
        """Learn a discretisation model using UMODL.

        Parameters
        ----------
        data: pd.DataFrame
            Dataframe containing feature variables. Categorical
            variables should have the object dtype, otherwise they
            are processed as numerical variables.
        treatment_col: pd.Series
            Treatment column.
        target_col: pd.Series
            Outcome column.
        maxpartnumber: int, default=None
            The maximal number of intervals or groups. None means default to the 'umodl' program default.
        """
        # Force the types of the treatment and target columns so that khiops lib understands they are categorical
        if treatment_col.dtype != object:
            warn("Treatment column's dtype is not object; fixing...")
            treatment_col = treatment_col.astype(object)
        if target_col.dtype != object:
            warn("Target column's dtype is not object; fixing...")
            target_col = target_col.astype(object)

        # Create temporary .txt (data) and .kdic (Khiops dictionary) files for use with the umodl executable
        with tempfile.TemporaryDirectory() as dirname:
            dirpath = pathlib.Path(dirname)
            kdicfilename = str(dirpath / "main_table.kdic")
            dataset = khiops.sklearn.dataset.Dataset(data.join([treatment_col, target_col]))

            txtfilename, _ = dataset.create_table_files_for_khiops(dirname)  # Create .txt file
            dataset.create_khiops_dictionary_domain().export_khiops_dictionary_file(kdicfilename)  # Create .kdic file
            run_umodl(txtfilename, kdicfilename, "main_table", treatment_col.name, target_col.name, maxpartnumber)

            txtfilepath = pathlib.Path(txtfilename)
            with open(txtfilepath.with_name(f"UP_{txtfilepath.stem}.json")) as jsonfile:
                docroot = json.load(jsonfile)

        model = {}
        for variable in docroot['detailed statistics']:
            vardim = variable['dataGrid']['dimensions'][0]
            if vardim['partitionType'] == 'Value groups':
                model[vardim['variable']] = ValGrpPartition(list(map(ValGrp, vardim['partition'])), vardim['defaultGroupIndex'])
            elif vardim['partitionType'] == 'Intervals':
                model[vardim['variable']] = IntervalPartition(list(starmap(Interval, vardim['partition'])))
            else: raise ValueError("unsupported partition type")

        levels = sorted(
            ((attr['name'], attr['level']) for attr in docroot['attributes']),
            key=lambda namelevel: (-namelevel[1], namelevel[0])
        )

        # Only write to the instance's attributes if all the above succeeded.
        self.model = model
        self.levels = levels
        self.variable_cols = data
        self.treatment_col = treatment_col
        self.target_col = target_col


    def transform(self, data):
        """Apply the discretisation model learned by the fit() method.

        Parameters
        ----------
        data: pd.DataFrame
            Dataframe containing feature variables.

        Returns
        -------
        pd.DataFrame
            Pandas Dataframe that contains encoded data.
        """
        data = data[list(self.model.keys())]  # Keep only informative variables
        self.transformed_data = data.transform(lambda col: self.model[col.name].transform(col))
        return self.transformed_data
    
    def get_levels(self):
        """Get the level of all variables.
        
        Returns
        -------
        list[tuple[str, float]]
            (variable-name, variable-level) pairs in decreasing level order.
        """
        return self.levels
    
    def get_partition(self, variable):
        """Get the partition corresponding to a single variable of the model.

        Parameters
        ----------
        variable: str
            The variable name.
        
        Returns
        -------
        ValGrpPartition | IntervalPartition
            The partition corresponding to a single variable of the model.
        """
        return self.model[variable]
    
    def get_target_frequencies(self, variable):
        """Get the frequencies for each (target, treatment) pair.
        
        The frequencies are computed for a single variable.
        
        Parameters
        ----------
        variable: str
            The variable name.

        Returns
        -------
        pd.DataFrame
            The frequencies as a Dataframe containing:
                - A column named 'Part' listing all the parts of the variable.
                - One column per (target, treatment) pair.
        """
        varcol = self.variable_cols[variable]
        partition = self.get_partition(variable)
        return pd.DataFrame(
            {
                **{"Part": partition},
                **{
                    ttpair: [
                        len(
                            varcol[
                                (self.treatment_col == ttpair.treatment) & (self.target_col == ttpair.target) & varcol.map(lambda elem: partition.transform_elem(elem) == i)
                            ]
                        )
                        for i, _ in enumerate(partition)
                    ]
                    for ttpair in self.target_treatment_pairs
                }
            }
        )
    
    def get_target_probabilities(self, variable):
        """Get the probabilities P(target|treatment) for each (target, treatment) pair.
        
        The probabilities are computed for a single variable.
        
        Parameters
        ----------
        variable: str
            The variable name.

        Returns
        -------
        pd.DataFrame
            The probabilities as a Dataframe containing:
                - A column named 'Part' listing all the parts of the variable.
                - One column per (target, treatment) pair.
        """
        return self.get_target_frequencies(variable).transform({
            "Part": lambda x: x,
            **{ttpair: lambda x: x / len(self.variable_cols) for ttpair in self.target_treatment_pairs}
        })
    
    def get_uplift(self, reftarget, reftreatment, variable):
        """Get the uplift for a single variable.

        See explanations of the computations in the 'Returns' section below.
        
        Parameters
        ----------
        reftarget
            The reference target.
        reftreatment
            The reference treatment to which all the other treatments are compared.
        variable: str
            The name of the variable.

        Returns
        -------
        pd.DataFrame
            A Dataframe containing:
                - A column named 'Part' listing all the parts of the variable.
                - One column per treatment other than the reference treatment.
                  A column gives the difference P(reftarget|treatment) - P(reftarget|reftreatment), that is,
                  the benefit (or deficit) of probabilities to have 'reftarget' as the outcome with the column's
                  treatment compared to the reference treatment.
        """
        probs = self.get_target_probabilities(variable)
        refprobs = probs[TargetTreatmentPair(reftarget, reftreatment)]
        return probs["Part"].to_frame().join(pd.DataFrame({
            f"Uplift {reftarget} {treatment}": probs[TargetTreatmentPair(reftarget, treatment)] - refprobs
            for treatment in self.treatment_modalities if treatment != reftreatment
        }))



######################################################################################
# Copyright (c) 2023 Orange - All Rights Reserved                             #
# * This software is the confidential and proprietary information of Orange.         #
# * You shall not disclose such Restricted Information and shall use it only in      #
#   accordance with the terms of the license agreement you entered into with Orange  #
#   named the "kuplift - Python Library Evaluation License".                          #
# * Unauthorized copying of this file, via any medium is strictly prohibited.        #
# * See the "LICENSE.md" file for more details.                                      #
######################################################################################
import numpy as np
import pandas as pd
from math import log
from .helperfunctions import (
    log_fact,
    universal_code_natural_numbers,
)
from .node import _Node


class _Tree:
    """Private parent class

    Parameters
    ----------
    data : pd.Dataframe
        Dataframe containing feature variables.
    treatment_col : pd.Series
        Treatment column.
    y_col : pd.Series
        Outcome column.
    """

    def __init__(self,control_name=None):  # ordered data as argument
        self.nodes_ids = None
        self.root_node = None
        self.terminal_nodes = []
        self.internal_nodes = []

        self.k = None
        self.k_t = None
        self.features = None
        self.feature_subset = None

        self.prob_kt = None
        self.encoding_of_being_an_internal_node = None
        self.prob_attribute_selection = None
        self.prior_of_internal_nodes = None
        self.encoding_of_being_a_leaf_node_and_containing_te = None
        self.leaf_prior = None
        self.tree_likelihood = None
        
        self.summary_df = None
        
        self.treatment_name='treatment'
        self.outcome_name='outcome'
        self.control_name=control_name
        
    def calc_criterion(self):
        self.__calc_prob_kt()
        self.__calc_prior_of_internal_node()
        self.__calc_encoding()
        self.__calc_leaf_prior()
        self.__calc_tree_likelihood()

    def __calc_prob_kt(self):
        self.prob_kt = (
            universal_code_natural_numbers(self.k_t)
            - log_fact(self.k_t)
            + self.k_t * log(self.k)
        )

    def __calc_prior_of_internal_node(self):
        if len(self.internal_nodes) == 0:
            self.prior_of_internal_nodes = 0
            self.prob_attribute_selection = 0
        else:
            prior_of_internal_nodes = 0
            for internal_node in self.internal_nodes:
                prior_of_internal_nodes += internal_node.prior_of_internal_node
            self.prior_of_internal_nodes = prior_of_internal_nodes
            self.prob_attribute_selection = log(self.k_t) * len(
                self.internal_nodes
            )

    def __calc_encoding(self):
        self.encoding_of_being_a_leaf_node_and_containing_te = (
            len(self.terminal_nodes) * log(2) * 2
        )
        self.encoding_of_being_an_internal_node = len(
            self.internal_nodes
        ) * log(2)

    def __calc_leaf_prior(self):
        leaf_priors = 0
        for leaf_node in self.terminal_nodes:
            leaf_priors += leaf_node.prior_leaf
        self.leaf_prior = leaf_priors

    def __calc_tree_likelihood(self):
        leaf_likelihoods = 0
        for leaf_node in self.terminal_nodes:
            leaf_likelihoods += leaf_node.likelihood_leaf
        self.tree_likelihood = leaf_likelihoods

    def __traverse_tree(self, x, node):
        if node.is_leaf:
            return node.average_uplift

        if x[node.attribute] <= node.split_threshold:
            return self.__traverse_tree(x, node.left_node)
        return self.__traverse_tree(x, node.right_node)
    
    def __initializeVars__(self, data, treatment_col, y_col):
        data = data.assign(**{self.treatment_name: treatment_col.copy()})
        data = data.assign(**{self.outcome_name: y_col.copy()})
        
        #dealing with control name
        if self.control_name != None:
            trt_vals=list(data[self.treatment_name].unique())
            
            #Verify that control name is in the treatment column, else raise an exception
            if self.control_name not in trt_vals:
                raise Exception("the control name is not in the treatment column")
            data[self.treatment_name] = data[self.treatment_name].replace(self.control_name,0)

            trt_vals.remove(self.control_name)
            #the other value will be considered as the treatment
            data[self.treatment_name] = data[self.treatment_name].replace(trt_vals[0],1)
        
        self.nodes_ids = 0
        self.root_node = _Node(
            data, self.treatment_name, self.outcome_name, ID=self.nodes_ids + 1
        )
        self.terminal_nodes = [self.root_node]
        self.internal_nodes = []

        self.k = len(list(data.columns))
        self.k_t = 1
        self.features = list(data.columns)
        self.feature_subset = []

        self.encoding_of_being_a_leaf_node_and_containing_te = (
            len(self.terminal_nodes) * log(2) * 2
        )  # TE=TreatmentEffect
        
        self.calc_criterion()

        self.tree_criterion = (
            self.prob_kt
            + self.encoding_of_being_an_internal_node
            + self.prob_attribute_selection
            + self.prior_of_internal_nodes
            + self.encoding_of_being_a_leaf_node_and_containing_te
            + self.leaf_prior
            + self.tree_likelihood
        )        

    def predict(self, X_test):
        """Predict the uplift value for each example in X_test

        Parameters
        ----------
        X_train : pd.Dataframe
            Dataframe containing feature variables.

        Returns
        -------
        y_pred_list(ndarray, shape=(num_samples, 1))
            An array containing the predicted treatment uplift for each sample.
        """
        predictions = [
            self.__traverse_tree(X_test.iloc[x], self.root_node)
            for x in range(len(X_test))
        ]
        return np.array(predictions)

    def getSummary(self):
        summary_df = pd.DataFrame(
            columns=[
                "NodeId",
                "is_leaf",
                "T0Y0",
                "T0Y1",
                "T1Y0",
                "T1Y1",
                "Uplift",
                "SplittedAttribute",
                "split_threshold",
            ]
        )  # split_threshold
        for internalNode in self.internal_nodes:
            summary_df.loc[len(summary_df.index)] = [
                internalNode.id,
                internalNode.is_leaf,
                internalNode.ntj[0],
                internalNode.ntj[1],
                internalNode.ntj[2],
                internalNode.ntj[3],
                internalNode.average_uplift,
                internalNode.attribute,
                internalNode.split_threshold,
            ]
        for terminalNode in self.terminal_nodes:
            summary_df.loc[len(summary_df.index)] = [
                terminalNode.id,
                terminalNode.is_leaf,
                terminalNode.ntj[0],
                terminalNode.ntj[1],
                terminalNode.ntj[2],
                terminalNode.ntj[3],
                terminalNode.average_uplift,
                terminalNode.attribute,
                terminalNode.split_threshold,
            ]
        self.summary_df = summary_df
        return self.summary_df

    def export_tree(self, IdValue=1, numTabs=0, text_desc=""):
        def create_tabs(txt, numTabs):
            for numTab in range(numTabs):
                txt += "\t"
            return txt

        # fill the summary dataframe of the tree
        if IdValue == 1:
            self.getSummary()

        row = (
            self.summary_df[self.summary_df["NodeId"] == IdValue]
            .iloc[:1]
            .reset_index(drop=True)
            .squeeze()
        )
        #     print("row is ",type(row))
        #     print("row is ",row)
        if row["is_leaf"] == False:
            #         print(" id ",str(IdValue)," not leaf")
            text_desc = create_tabs(text_desc, numTabs)
            text_desc = (
                text_desc
                + "|--- "
                + " "
                + str(row["SplittedAttribute"])
                + " <= "
                + str(row["split_threshold"])
                + "\n"
            )
            #         print(text_desc)
            text_desc = self.export_tree(IdValue * 2, numTabs + 1, text_desc)

            text_desc = create_tabs(text_desc, numTabs)
            text_desc = (
                text_desc
                + "|--- "
                + " "
                + str(row["SplittedAttribute"])
                + " > "
                + str(row["split_threshold"])
                + "\n"
            )
            text_desc = self.export_tree(
                IdValue * 2 + 1, numTabs + 1, text_desc
            )
        else:
            text_desc = create_tabs(text_desc, numTabs)
            text_desc += "|--- Leaf \n"
            text_desc = create_tabs(text_desc, numTabs + 1)
            try:
                text_desc = (
                    text_desc
                    + "|--- "
                    + " Outcome Distribution in Treatment "
                    + str(row["T1Y1"] / (row["T1Y1"] + row["T1Y0"]))
                    + "\n"
                )
            except:
                text_desc = (
                    text_desc
                    + "|--- "
                    + " Outcome Distribution in Treatment "
                    + str(row["T1Y1"] / 0.0001)
                    + "(No treatment)\n"
                )

            text_desc = create_tabs(text_desc, numTabs + 1)
            try:
                text_desc = (
                    text_desc
                    + "|--- "
                    + " Outcome Distribution in Control "
                    + str(row["T0Y1"] / (row["T0Y1"] + row["T0Y0"]))
                    + "\n"
                )
            except:
                text_desc = (
                    text_desc
                    + "|--- "
                    + " Outcome Distribution in Control "
                    + str(row["T0Y1"] / 0.001)
                    + " (No control)\n"
                )

        return text_desc



######################################################################################
# Copyright (c) 2023 Orange - All Rights Reserved                             #
# * This software is the confidential and proprietary information of Orange.         #
# * You shall not disclose such Restricted Information and shall use it only in      #
#   accordance with the terms of the license agreement you entered into with Orange  #
#   named the "kuplift - Python Library Evaluation License".                          #
# * Unauthorized copying of this file, via any medium is strictly prohibited.        #
# * See the "LICENSE.md" file for more details.                                      #
######################################################################################
import multiprocessing as mp
from .helperfunctions import preprocess_data
from .umodl_search_algorithm import execute_greedy_search_and_post_opt
import pandas as pd


class FeatureSelection:
    """
    The FeatureSelection implements the feature selection algorithm 'UMODL-FS'
    described in: Rafla, M., Voisine, N., Crémilleux, B., & Boullé, M.
    (2023, March). A non-parametric bayesian approach for uplift
    discretization and feature selection.
    In Machine Learning and Knowledge Discovery in Databases:
    European Conference, ECML PKDD 2022, Grenoble, France,
    September 19–23, 2022, Proceedings, Part V (pp. 239-254).
    Cham: Springer Nature Switzerland.
    """
    def __init__(self,control_name=None):
        self.treatment_name = "treatment"
        self.outcome_name = "outcome"
        self.control_name=control_name
        self.features_importance_details={}
        self.dict_vars_vs_importance={}
    
    def get_features_importance_details(self):
        '''
        After launch the feature selection approach, this function helps getting the details of a each feature. 
        How it was discretized, the intervals, the outcome denisities in each interval.
        '''
        
        # Prepare lists to capture multi-index levels
        outer_keys = []
        inner_keys = []
        values = []
        
        dictionary_of_dataframes={}
        # Loop through the dictionary to populate the lists
        for key, sublist in self.features_importance_details.items():
            for idx, val in enumerate(sublist):
                outer_keys.append(key)
                inner_keys.append(idx)
                values.append(val)
        
        # Create the multi-index
        multi_index = pd.MultiIndex.from_tuples(list(zip(outer_keys, inner_keys)), names=["Variable", "Interval"])
        
        # Convert the lists to DataFrame
        df = pd.DataFrame(values, columns=["interval", "P(Y|T=1)", "P(Y|T=0)", "uplift"], index=multi_index)
        
        df["Imp. Score"] = df.index.get_level_values(0).map(self.dict_vars_vs_importance)
        
        # Get sorted unique outer indices based on Imp. Score
        unique_outer_indices = df.groupby(level=0)['Imp. Score'].first().sort_values(ascending=False).index

        # Reindex the dataframe to rearrange based on the sorted outer indices
        df_sorted = df.reindex(unique_outer_indices, level=0)
        
        return df_sorted
    
    def __get_the_best_var(self, data, treatment_col, y_col):
        """
        Parameters
        ----------
        data : pd.Dataframe
            Dataframe containing data.
        treatment_col : pd.Series
            Treatment column.
        y_col : pd.Series
            Outcome column.

        Returns
        -------
        dict
            A Python dictionary containing the sorted variable importance,
            where the keys represent the variable names and the values denote
            their respective importance.

        For example: return a dictionary
                    var_vs_importance={"age":2.2,"job":2.3}
        """
        features = list(data.columns)
        features.remove(self.treatment_name)
        features.remove(self.outcome_name)

        var_vs_importance = {}
        var_vs_disc = {}
        for feature in features:
            (
                var_vs_importance[feature],
                var_vs_disc[feature],
                col_name,
                self.features_importance_details[feature]
            ) = execute_greedy_search_and_post_opt(
                data[[feature, self.treatment_name, self.outcome_name]],get_intervals_uplift=True
            )
        # sort the dictionary by values in ascending order
        var_vs_importance = {
            k: v
            for k, v in sorted(
                var_vs_importance.items(), key=lambda item: item[1]
            )
        }
        return var_vs_importance

    def filter(
        self, data, treatment_col, y_col, parallelized=False, num_processes=5
    ):
        """
        This function runs the feature selection algorithm 'UMODL-FS',
        ranking variables based on their importance in the given data.

        Parameters
        ----------
        data : pd.Dataframe
            Dataframe containing feature variables.
        treatment_col : pd.Series
            Treatment column.
        y_col : pd.Series
            Outcome column.
        parallelized : bool, default False
            Whether to run the code on several processes.
        num_processes : int, default 5
            Number of processes to use in parallel, 'parallelized' argument should be True.

        Returns
        -------
        Python Dictionary
            Variables names and their corresponding importance value (Sorted).
        """
        data = data.assign(**{self.treatment_name: treatment_col.copy()})
        data = data.assign(**{self.outcome_name: y_col.copy()})
        
        #dealing with control name
        if self.control_name != None:
            trt_vals=list(data[self.treatment_name].unique())
            
            #Verify that control name is in the treatment column, else raise an exception
            if self.control_name not in trt_vals:
                raise Exception("the control name is not in the treatment column")
            data[self.treatment_name] = data[self.treatment_name].replace(self.control_name,0)

            trt_vals.remove(self.control_name)
            #the other value will be considered as the treatment
            data[self.treatment_name] = data[self.treatment_name].replace(trt_vals[0],1)

        
        cols = list(data.columns)
        cols.remove(self.treatment_name)
        cols.remove(self.outcome_name)

        data = data[cols + [self.treatment_name, self.outcome_name]]
        data = preprocess_data(data, self.treatment_name, self.outcome_name)
        
        self.var_vs_disc = {}
        
        if parallelized:
            pool = mp.Pool(processes=num_processes)

            arguments_to_pass_in_parallel = []
            for col in cols:
                arguments_to_pass_in_parallel.append(
                    [data[[col, self.treatment_name, self.outcome_name]],True]
                )
            list_of_tuples_feature_vs_importance = pool.starmap(
                execute_greedy_search_and_post_opt,
                arguments_to_pass_in_parallel
            )
            pool.close()

            for el in list_of_tuples_feature_vs_importance:
                col = el[2]
                self.features_importance_details[col]=el[3]
                if len(el[1]) == 1:
                    self.var_vs_disc[col] = 0
                else:
                    self.var_vs_disc[col] = el[0]
            self.var_vs_disc = {
                k: v
                for k, v in sorted(
                    self.var_vs_disc.items(), key=lambda item: item[1]
                )
            }
            self.dict_vars_vs_importance=self.var_vs_disc.copy()
            return self.var_vs_disc
            

        else:
            list_of_vars_importance = self.__get_the_best_var(
                data, self.treatment_name, self.outcome_name
            )
            self.dict_vars_vs_importance=list_of_vars_importance.copy()
        
        return list_of_vars_importance



######################################################################################
# Copyright (c) 2023 Orange - All Rights Reserved                             #
# * This software is the confidential and proprietary information of Orange.         #
# * You shall not disclose such Restricted Information and shall use it only in      #
#   accordance with the terms of the license agreement you entered into with Orange  #
#   named the "kuplift - Python Library Evaluation License".                          #
# * Unauthorized copying of this file, via any medium is strictly prohibited.        #
# * See the "LICENSE.md" file for more details.                                      #
######################################################################################
from .binary_discretization_functions import umodl_binary_discretization
from .helperfunctions import (
    log_fact,
    log_binomial_coefficient,
)


class _Node:
    """Private class

    Parameters
    ----------
    data : pd.Dataframe
        Dataframe containing feature variables.
    treatment_col : pd.Series
        Treatment column.
    y_col : pd.Series
        Outcome column.
    ID : int, default None
        Tree node ID.
    """

    def __init__(self, data, treatment_col, y_col, ID=None):
        # Initialize attributes
        self.id = ID
        self.treatment = treatment_col
        self.output = y_col
        self.n = data.shape[0]
        self.nj = data[data[self.output] == 1].shape[0]
        self.ntj = [
            data[(data[self.treatment] == 0) & (data[self.output] == 0)].shape[
                0
            ],
            data[(data[self.treatment] == 0) & (data[self.output] == 1)].shape[
                0
            ],
            data[(data[self.treatment] == 1) & (data[self.output] == 0)].shape[
                0
            ],
            data[(data[self.treatment] == 1) & (data[self.output] == 1)].shape[
                0
            ],
        ]
        self.x = data.iloc[:, :-2].copy()
        self.t = data.iloc[:, -2].copy()
        self.y = data.iloc[:, -1].copy()

        try:
            if (self.ntj[2] + self.ntj[3]) == 0:
                denum = 0.00001
            else:
                denum = self.ntj[2] + self.ntj[3]
            self.outcome_prob_in_trt = self.ntj[3] / denum
        except Exception:
            self.outcome_prob_in_trt = 0

        try:
            if (self.ntj[0] + self.ntj[1]) == 0:
                denum = 0.00001
            else:
                denum = self.ntj[0] + self.ntj[1]
            self.outcome_prob_in_ctrl = self.ntj[1] / denum
        except Exception:
            self.outcome_prob_in_ctrl = 0

        self.average_uplift = (
            self.outcome_prob_in_trt - self.outcome_prob_in_ctrl
        )
        self.attribute = None
        self.split_threshold = None
        self.is_leaf = True
        self.candidate_splits_vs_data_left_data_right = None
        self.candidate_splits_vs_criterion = None
        self.left_node = None
        self.right_node = None
        self.prior_of_internal_node = self.__calc_prior_of_internal_node()
        (
            self.prior_leaf,
            self.likelihood_leaf,
            self.w,
        ) = self.__calc_prior_and_likelihood_leaf()

    def __calc_prior_of_internal_node(self):
        return log_binomial_coefficient(sum(self.ntj) + 1, 1)

    def __calc_prior_and_likelihood_leaf(self):
        number_of_treatment = self.ntj[2] + self.ntj[3]
        number_of_control = self.ntj[0] + self.ntj[1]
        number_of_pos_outcome = self.ntj[1] + self.ntj[3]
        number_of_zero_outcome = self.ntj[0] + self.ntj[2]

        leaf_prior_w_zero = log_binomial_coefficient(sum(self.ntj) + 1, 1)
        tree_likelihood_w_zero = (
            log_fact(sum(self.ntj))
            - log_fact(number_of_pos_outcome)
            - log_fact(number_of_zero_outcome)
        )

        leaf_prior_w_one = log_binomial_coefficient(
            number_of_treatment + 1, 1
        ) + log_binomial_coefficient(number_of_control + 1, 1)
        tree_likelihood_w_one = (
            log_fact(number_of_treatment)
            - log_fact(self.ntj[2])
            - log_fact(self.ntj[3])
        ) + (
            log_fact(number_of_control)
            - log_fact(self.ntj[0])
            - log_fact(self.ntj[1])
        )

        if (leaf_prior_w_zero + tree_likelihood_w_zero) < (
            leaf_prior_w_one + tree_likelihood_w_one
        ):
            w = 0
            leaf_prior = leaf_prior_w_zero
            tree_likelihood = tree_likelihood_w_zero
        else:
            w = 1
            leaf_prior = leaf_prior_w_one
            tree_likelihood = tree_likelihood_w_one
        return leaf_prior, tree_likelihood, w

    def discretize_vars_and_get_attributes_splits_costs(self):
        """For this node loop on all attributes
        and get the optimal split for each one.

        Returns
        -------
        Dictionary of lists

        For example: return a dictionnary {age: Cost, sex: Cost}
        The cost here corresponds to
        1- the cost of this node to internal instead of leaf
            (criterion_to_be_internal-prior_leaf)
        2- The combinatorial terms of the leaf prior and likelihood
        """
        features = list(self.x.columns)
        attribute_to_split_vs_left_and_right_data = {}
        for attribute in features:
            if (
                len(self.x[attribute].value_counts()) == 1
                or len(self.x[attribute].value_counts()) == 0
            ):
                continue
            disc_res = umodl_binary_discretization(
                self.x, self.t, self.y, attribute
            )
            if disc_res == -1:
                continue
            data_left, data_right, threshold = (
                disc_res[0],
                disc_res[1],
                disc_res[2],
            )
            attribute_to_split_vs_left_and_right_data[attribute] = [
                data_left,
                data_right,
                threshold,
            ]

        self.candidate_splits_vs_data_left_data_right = (
            attribute_to_split_vs_left_and_right_data.copy()
        )
        candidate_splits_vs_criterion = self.__get_attributes_splits_costs(
            attribute_to_split_vs_left_and_right_data
        )
        self.candidate_splits_vs_criterion = (
            candidate_splits_vs_criterion.copy()
        )
        return candidate_splits_vs_criterion.copy()

    def __get_attributes_splits_costs(self, dict_of_each_att_vs_effectifs):
        # Prior of Internal node is only the combinatorial calculations
        # In case we split this node, it will be no more a leaf but an internal
        # node
        criterion_to_be_internal = self.__calc_prior_of_internal_node()
        new_prior_vals = (
            criterion_to_be_internal - self.prior_leaf - self.likelihood_leaf
        )

        candidate_splits_vs_criterion = {}
        for key in dict_of_each_att_vs_effectifs:
            leaves_val = self.__update_tree_criterion(
                dict_of_each_att_vs_effectifs[key][:2]
            )
            candidate_splits_vs_criterion[key] = new_prior_vals + leaves_val
        return candidate_splits_vs_criterion.copy()

    def __update_tree_criterion(self, left_and_right_data):
        leaves_vals = 0
        for (
            new_node_effectifs
        ) in left_and_right_data:  # Loop on Left and Right candidate nodes
            node = _Node(new_node_effectifs, self.treatment, self.output)
            leaves_vals += node.prior_leaf + node.likelihood_leaf
            del node
        return leaves_vals

    def perform_split(self, attribute):
        if self.candidate_splits_vs_data_left_data_right is None:
            raise
        else:
            self.is_leaf = False
            self.left_node = _Node(
                self.candidate_splits_vs_data_left_data_right[attribute][0],
                self.treatment,
                self.output,
                ID=self.id * 2,
            )
            self.right_node = _Node(
                self.candidate_splits_vs_data_left_data_right[attribute][1],
                self.treatment,
                self.output,
                ID=self.id * 2 + 1,
            )
            self.attribute = attribute
            self.split_threshold = (
                self.candidate_splits_vs_data_left_data_right[attribute][2]
            )
            del self.x
            del self.t
            del self.y
            del self.candidate_splits_vs_data_left_data_right

            return self.left_node, self.right_node



######################################################################################
# Copyright (c) 2023 Orange - All Rights Reserved                             #
# * This software is the confidential and proprietary information of Orange.         #
# * You shall not disclose such Restricted Information and shall use it only in      #
#   accordance with the terms of the license agreement you entered into with Orange  #
#   named the "kuplift - Python Library Evaluation License".                          #
# * Unauthorized copying of this file, via any medium is strictly prohibited.        #
# * See the "LICENSE.md" file for more details.                                      #
######################################################################################
from math import log
from .helperfunctions import (
    log_fact,
    universal_code_natural_numbers,
)
from .tree import _Tree


class BayesianDecisionTree(_Tree):
    """
    The BayesianDecisionTree class implements the UB-DT algorithm described in:
    Rafla, M., Voisine, N., Crémilleux, B., & Boullé, M. (2023, May).
    A Non-Parametric Bayesian Decision Trees for Uplift modelling. In PAKDD.

    Parameters
    ----------
    data : pd.Dataframe
        Dataframe containing feature variables.
    treatment_col : pd.Series
        Treatment column.
    y_col : pd.Series
        Outcome column.
    control_name: int or str
        The name of the control value in the treatment column
    """

    def __init__(self, control_name=None):
        super().__init__(control_name)

    def fit(self,data, treatment_col, y_col):
        """Fit an uplift decision tree model using UB-DT

        Parameters
        ----------
        X_train : pd.Dataframe
            Dataframe containing feature variables.
        treatment_col : pd.Series
            Treatment column.
        y_col : pd.Series
            Outcome column.
        """
        super().__initializeVars__(data, treatment_col, y_col)
        # In case if we have a new attribute for splitting
        prob_kt_plus_one = (
            universal_code_natural_numbers(self.k_t + 1)
            - log_fact(self.k_t + 1)
            + (self.k_t + 1) * log(self.k)
        )
        prob_of_attribute_selection_among_subset_attributes_plus_one = log(
            self.k_t + 1
        ) * (len(self.internal_nodes) + 1)

        encoding_of_being_an_internal_node_plus_one = (
            self.encoding_of_being_an_internal_node + log(2)
        )

        # When splitting a node to 2 nodes, the number of leaf nodes is
        # incremented only by one, since the parent node was leaf
        # and is now internal.
        # 2 for two extra leaf nodes multiplied by 2 for w. Total = 4.
        encoding_of_being_a_leaf_node_and_containing_te_plus_two = (
            self.encoding_of_being_a_leaf_node_and_containing_te + (2 * log(2))
        )

        encoding_of_internal_and_leaves_and_w_with_extra_nodes = (
            encoding_of_being_an_internal_node_plus_one
            + encoding_of_being_a_leaf_node_and_containing_te_plus_two
        )

        i = 0
        while True:
            node_vs_best_attribute_corresponding_to_the_best_cost = {}
            node_vs_best_cost = {}
            node_vs_candidate_splits_costs = (
                {}
            )  # Dictionary containing Nodes as key and their values are another dictionary each with attribute:CostSplit

            for terminal_node in self.terminal_nodes:
                # This if condition is here to not to repeat calculations of candidate splits
                if terminal_node.candidate_splits_vs_criterion is None:
                    node_vs_candidate_splits_costs[
                        terminal_node
                    ] = (
                        terminal_node.discretize_vars_and_get_attributes_splits_costs()
                    )
                else:
                    node_vs_candidate_splits_costs[
                        terminal_node
                    ] = terminal_node.candidate_splits_vs_criterion.copy()

                if len(node_vs_candidate_splits_costs[terminal_node]) == 0:
                    continue

                # Update Costs
                for attribute in node_vs_candidate_splits_costs[terminal_node]:
                    if attribute in self.feature_subset:
                        node_vs_candidate_splits_costs[terminal_node][
                            attribute
                        ] += (
                            self.prob_kt
                            + self.prob_attribute_selection
                            + encoding_of_internal_and_leaves_and_w_with_extra_nodes
                            + self.leaf_prior
                            + self.tree_likelihood
                            + self.prior_of_internal_nodes
                        )
                    else:
                        node_vs_candidate_splits_costs[terminal_node][
                            attribute
                        ] += (
                            prob_kt_plus_one
                            + encoding_of_internal_and_leaves_and_w_with_extra_nodes
                            + prob_of_attribute_selection_among_subset_attributes_plus_one
                            + self.leaf_prior
                            + self.tree_likelihood
                            + self.prior_of_internal_nodes
                        )

                # Once costs are updated, I get the key of the minimal value
                # split for terminal_node
                key_of_the_minimal_val = min(
                    node_vs_candidate_splits_costs[terminal_node],
                    key=node_vs_candidate_splits_costs[terminal_node].get,
                )

                node_vs_best_attribute_corresponding_to_the_best_cost[
                    terminal_node
                ] = key_of_the_minimal_val
                node_vs_best_cost[
                    terminal_node
                ] = node_vs_candidate_splits_costs[terminal_node][
                    key_of_the_minimal_val
                ]

            if len(list(node_vs_best_cost)) == 0:
                break

            optimal_node_attribute_to_split_up = min(
                node_vs_best_cost, key=node_vs_best_cost.get
            )
            optimal_val = node_vs_best_cost[optimal_node_attribute_to_split_up]
            optimal_node = optimal_node_attribute_to_split_up
            optimal_attribute = (
                node_vs_best_attribute_corresponding_to_the_best_cost[
                    optimal_node_attribute_to_split_up
                ]
            )

            if optimal_val < self.tree_criterion:
                self.tree_criterion = optimal_val
                if optimal_attribute not in self.feature_subset:
                    self.feature_subset.append(optimal_attribute)
                    self.k_t += 1
                new_left_leaf, new_right_leaf = optimal_node.perform_split(
                    optimal_attribute
                )
                self.terminal_nodes.append(new_left_leaf)
                self.terminal_nodes.append(new_right_leaf)
                self.internal_nodes.append(optimal_node)
                self.terminal_nodes.remove(optimal_node)

                self.calc_criterion()
            else:
                break



######################################################################################
# Copyright (c) 2023 Orange - All Rights Reserved                             #
# * This software is the confidential and proprietary information of Orange.         #
# * You shall not disclose such Restricted Information and shall use it only in      #
#   accordance with the terms of the license agreement you entered into with Orange  #
#   named the "kuplift - Python Library Evaluation License".                          #
# * Unauthorized copying of this file, via any medium is strictly prohibited.        #
# * See the "LICENSE.md" file for more details.                                      #
######################################################################################
import numpy as np
from math import log
from operator import add, sub, itemgetter
from sortedcontainers import SortedKeyList
from .helperfunctions import log_fact, log_binomial_coefficient
from .binary_discretization_functions import start_counter, stop_counter


class _Interval:
    def __init__(self, data):
        if len(data) == 3:
            self.nitj = data[0]
            self.included_right_frontier = data[1]
            self.w = data[2]
            self.right_merge_delta = None
            self.right_merge_w = None
            self.prior_one = None
            self.prior_two = None
            self.likelihood_one = None
            self.likelihood_two = None
            self.next = None
            self.previous = None
            self.sum_of_priors_and_likelihoods = None
            self.sum_of_priors_and_likelihoods_with_zero_w = None
            self.sum_of_priors_and_likelihoods_with_one_w = None
        else:
            self.nitj = data[0]
            self.included_right_frontier = data[1]
            self.w = data[2]
            self.right_merge_delta = data[3]
            self.right_merge_w = data[4]
            self.prior_one = data[5]
            self.prior_two = data[6]
            self.likelihood_one = data[7]
            self.likelihood_two = data[8]
            self.next = data[9]
            self.previous = data[10]
            self.sum_of_priors_and_likelihoods = data[11]
            self.sum_of_priors_and_likelihoods_with_zero_w = None

    def get_interval_data(self):
        return [
            self.nitj,
            self.included_right_frontier,
            self.w,
            self.right_merge_delta,
            self.right_merge_w,
            self.prior_one,
            self.prior_two,
            self.likelihood_one,
            self.likelihood_two,
            self.next,
            self.previous,
            self.sum_of_priors_and_likelihoods,
        ]

    def calculate_priors_and_likelihoods(
        self, mode="init", calc_null_one=False, calc_null_two=False
    ):
        if mode == "DeltaCalc":
            interval_one_to_be_merged = self.nitj.copy()
            interval_two_to_be_merged = self.next.nitj.copy()
            nitj_interval = list(
                map(add, interval_one_to_be_merged, interval_two_to_be_merged)
            )
        else:
            nitj_interval = self.nitj.copy()

        # Likelihood 1
        likelihood_denum = 0
        for j in range(2):
            likelihood_denum += log_fact(
                (nitj_interval[j] + nitj_interval[j + 2])
            )
        likelihood_one_tmp = log_fact(np.sum(nitj_interval)) - likelihood_denum
        # Likelihood 2
        res_t = 0
        for t in range(2):
            likelihood_denum = 0
            for j in range(2):
                likelihood_denum += log_fact(nitj_interval[t + t * 1 + j])
            res_t += (
                log_fact(
                    nitj_interval[t + t * 1] + nitj_interval[t + t * 1 + 1]
                )
                - likelihood_denum
            )
        likelihood_two_tmp = res_t
        # Prior 1
        prior_one_tmp = log_binomial_coefficient(np.sum(nitj_interval) + 1, 1)
        # Prior 2
        res_t = 0
        for t in range(2):
            res_t_temp = log_binomial_coefficient(
                (nitj_interval[t + t * 1] + nitj_interval[t + t * 1 + 1]) + 1,
                1,
            )
            res_t += res_t_temp
        prior_two_tmp = res_t

        if mode == "DeltaCalc":
            if (prior_one_tmp + likelihood_one_tmp) < (
                prior_two_tmp + likelihood_two_tmp
            ):
                self.right_merge_w = 0
            else:
                self.right_merge_w = 1
            sum_of_priors_and_likelihoods = (
                (1 - self.right_merge_w) * prior_one_tmp
                + self.right_merge_w * prior_two_tmp
                + (1 - self.right_merge_w) * likelihood_one_tmp
                + (self.right_merge_w) * likelihood_two_tmp
            )

        elif mode == "MergeAndUpdate":
            if (
                (
                    (prior_one_tmp + likelihood_one_tmp)
                    < (prior_two_tmp + likelihood_two_tmp)
                )
                or calc_null_one
                or calc_null_two
            ):
                if calc_null_two:
                    self.right_merge_w = 1
                else:
                    self.right_merge_w = 0
            else:
                self.right_merge_w = 1
            self.prior_one = (1 - self.right_merge_w) * prior_one_tmp
            self.prior_two = self.right_merge_w * prior_two_tmp
            self.likelihood_one = (1 - self.right_merge_w) * likelihood_one_tmp
            self.likelihood_two = (self.right_merge_w) * likelihood_two_tmp
            self.w = self.right_merge_w
            sum_of_priors_and_likelihoods = (
                self.prior_one
                + self.prior_two
                + self.likelihood_one
                + self.likelihood_two
            )
            if calc_null_one:
                self.sum_of_priors_and_likelihoods_with_zero_w = (
                    sum_of_priors_and_likelihoods
                )
            if calc_null_two:
                self.sum_of_priors_and_likelihoods_with_one_w = (
                    sum_of_priors_and_likelihoods
                )
            else:
                self.sum_of_priors_and_likelihoods = (
                    sum_of_priors_and_likelihoods
                )

        elif mode == "init":
            sum_of_priors_and_likelihoods = (
                ((1 - self.w) * prior_one_tmp)
                + (self.w * prior_two_tmp)
                + ((1 - self.w) * likelihood_one_tmp)
                + ((self.w) * likelihood_two_tmp)
            )
            self.sum_of_priors_and_likelihoods = sum_of_priors_and_likelihoods
            self.prior_one = (1 - self.w) * prior_one_tmp
            self.prior_two = self.w * prior_two_tmp
            self.likelihood_one = (1 - self.w) * likelihood_one_tmp
            self.likelihood_two = (self.w) * likelihood_two_tmp
        return sum_of_priors_and_likelihoods


class _DLL:
    def __init__(self):
        self.head = None
        self.tail = None
        self.n = None
        self.modl_value = None
        self.count = 0
        self.i = self.count

    # append to the end of the list
    def append(self, listData):
        if self.head is None:
            self.head = _Interval(listData)
            self.tail = self.head
            self.count += 1
            self.i += 1
            return
        self.tail.next = _Interval(listData)
        self.tail.next.previous = self.tail
        self.tail = self.tail.next
        self.count += 1
        self.i += 1

    def insert(self, interval, index):
        nitj = interval.nitj
        included_right_frontier = interval.included_right_frontier
        W_value = interval.w

        if (index > self.count) | (index < 0):
            raise ValueError(
                f"Index out of range: {index}, size: {self.count}"
            )

        if index == self.count:
            self.append([nitj, included_right_frontier, W_value])
            return

        if index == 0:
            self.head.previous = _Interval(
                [nitj, included_right_frontier, W_value]
            )
            self.head.previous.next = self.head
            self.head = self.head.previous
            self.count += 1
            self.i += 1
            return

        start = self.head
        for _ in range(index):
            start = start.next
        start.previous.next = _Interval(
            [nitj, included_right_frontier, W_value]
        )
        start.previous.next.previous = start.previous
        start.previous.next.next = start
        start.previous = start.previous.next
        self.count += 1
        self.i += 1
        return

    def remove_interval(self, interval):
        if interval == self.head:
            self.head = self.head.next
            self.head.previous = None
            self.count -= 1
            self.i -= 1
            return

        if interval == self.tail:
            self.tail = self.tail.previous
            self.tail.next = None
            self.count -= 1
            self.i -= 1
            return

        interval.previous.next, interval.next.previous = (
            interval.next,
            interval.previous,
        )
        self.count -= 1
        self.i -= 1
        return

    def remove(self, index):
        if (index >= self.count) | (index < 0):
            raise ValueError(
                f"Index out of range: {index}, size: {self.count}"
            )

        if index == 0:
            self.head = self.head.next
            self.head.previous = None
            self.count -= 1
            self.i -= 1
            return

        if index == (self.count - 1):
            self.tail = self.tail.previous
            self.tail.next = None
            self.count -= 1
            self.i -= 1
            return

        start = self.head
        for i in range(index):
            start = start.next
        start.previous.next, start.next.previous = start.next, start.previous
        self.count -= 1
        self.i -= 1
        return

    def size(self):
        return self.count

    def get_nth(self, index):
        current = self.head  # Initialise temp
        count = 0  # Index of current interval

        # Loop while end of linked list is not reached
        while current:
            if count == index:
                return current
            count += 1
            current = current.next
        return 0

    def get_discretization_criterion_value(self):
        start = self.head
        summations = 0
        while start:
            summations += start.sum_of_priors_and_likelihoods
            start = start.next
        summations = (
            summations
            + log_binomial_coefficient(self.n + self.count - 1, self.count - 1)
            + self.count * log(2)
            + log(self.n)
        )
        self.modl_value = summations
        return summations

    def get_sorted_list_of_address_and_right_merge_value(self, interval):
        address_and_val = []  # list of lists
        while interval:
            rightMergeVal = interval.right_merge_delta
            address_and_val.append((rightMergeVal, interval))
            interval = interval.next
        address_and_val = sorted(address_and_val, key=itemgetter(0))
        return sorted(address_and_val, key=itemgetter(0))

    def get_discretization_info(self):
        start = self.head
        interval_bounds = []
        list_of_ws = []
        while start:
            interval_bounds.append(start.included_right_frontier)
            list_of_ws.append(start.w)
            start = start.next

        return [self.modl_value, self.count, interval_bounds, list_of_ws]


def create_elementary_discretization(dll, data):
    """
    params
    data : list of lists, each internal list contains [data value, treatment value, y value]

    returns
    1- a list of lists. Each internal list represents an interval and contains Effectifs of T0J0, T0J1, T1J0, T1J1 respectively.
    2- frontier value per interval
    3- w : list containing the Wi value for each interval, the initial discretization has Wi = 0 for all i
    """
    start_counter(0)
    # This is a list of lists, each internal list represents an interval and contains Effectifs of T0J0, T0J1, T1J0, T1J1 respectively
    prev = None
    i = -1
    for interval_list in data:
        if interval_list[0] != prev:
            if i != -1:
                dll.tail.calculate_priors_and_likelihoods()
            dll.append([[0, 0, 0, 0], interval_list[0], 0])
            prev = interval_list[0]
            i += 1
        dll.tail.nitj[int((interval_list[1] * 2) + interval_list[2])] += 1
    dll.tail.calculate_priors_and_likelihoods()
    stop_counter(0)

    # calculating whole criterion
    return dll, dll.get_discretization_criterion_value(), dll.size()


def criterion_delta_for_one_adjacent_interval_merge(
    dll, left_interval_node, index_passed=True, mode="DeltaCalc"
):
    start_counter(1)
    if index_passed:
        left_interval_node = dll.get_nth(left_interval_node)

    right_interval_node = left_interval_node.next
    if right_interval_node is None:
        # it means the left interval node is the last node and cannot be merged
        left_interval_node.right_merge_delta = -1
        return
    old_left_interval_node_criterion = (
        left_interval_node.sum_of_priors_and_likelihoods
    )

    left_interval_node_criterion = (
        left_interval_node.calculate_priors_and_likelihoods(mode)
    )

    right_interval_node_criterion = (
        right_interval_node.sum_of_priors_and_likelihoods
    )

    new_criterion_value = (
        dll.modl_value
        - right_interval_node_criterion
        - old_left_interval_node_criterion
        - log_binomial_coefficient(dll.n + dll.i - 1, dll.i - 1)
        - ((dll.i) * log(2))
        + log_binomial_coefficient(dll.n + dll.i - 2, dll.i - 2)
        + ((dll.i - 1) * log(2))
        + left_interval_node_criterion
    )

    left_interval_node.right_merge_delta = dll.modl_value - new_criterion_value
    stop_counter(1)


def compute_criterion_delta_for_all_possible_merges(dll):
    start_counter(2)
    size = dll.size()
    for i in range(size):
        criterion_delta_for_one_adjacent_interval_merge(dll, i)
    stop_counter(2)


def greedy_search(best_merges, intervals, N):
    start_counter(3)
    for step in range(N):
        if len(best_merges) > 0:
            best_merge_tuple = best_merges.pop()
        else:
            break
        if best_merge_tuple[0] >= 0:
            interval_to_be_merged = best_merge_tuple[1]
            interval_right_of_the_merge = interval_to_be_merged.next

            interval_one_to_be_merged = interval_to_be_merged.nitj.copy()
            interval_two_to_be_merged = interval_right_of_the_merge.nitj.copy()
            start_counter(6)

            merged_intervals = list(
                map(add, interval_one_to_be_merged, interval_two_to_be_merged)
            )
            stop_counter(6)

            interval_to_be_merged.nitj = merged_intervals
            interval_to_be_merged.included_right_frontier = (
                interval_right_of_the_merge.included_right_frontier
            )

            old_left_interval_node_criterion = (
                interval_to_be_merged.sum_of_priors_and_likelihoods
            )
            old_right_interval_node_criterion = (
                interval_right_of_the_merge.sum_of_priors_and_likelihoods
            )

            left_interval_node_criterion = interval_to_be_merged.calculate_priors_and_likelihoods(
                mode="MergeAndUpdate"
            )  # it will update w, Priors, lkelihoods and sum_of_priors_and_likelihoods
            intervals.modl_value = (
                intervals.modl_value
                - old_right_interval_node_criterion
                - old_left_interval_node_criterion
                - log_binomial_coefficient(
                    intervals.n + intervals.i - 1, intervals.i - 1
                )
                - ((intervals.i) * log(2))
                + log_binomial_coefficient(
                    intervals.n + intervals.i - 2, intervals.i - 2
                )
                + ((intervals.i - 1) * log(2))
                + left_interval_node_criterion
            )

            interval_right_to_new_interval = interval_to_be_merged.next
            interval_left_to_new_interval = interval_to_be_merged.previous

            best_merges.remove(
                (
                    interval_right_of_the_merge.right_merge_delta,
                    interval_right_of_the_merge,
                )
            )
            intervals.remove_interval(interval_right_of_the_merge)

            if interval_right_to_new_interval is None:  # last Interval
                best_merges.remove(
                    (
                        interval_left_to_new_interval.right_merge_delta,
                        interval_left_to_new_interval,
                    )
                )
                criterion_delta_for_one_adjacent_interval_merge(
                    intervals,
                    interval_left_to_new_interval,
                    index_passed=False,
                )
                best_merges.add(
                    (
                        interval_left_to_new_interval.right_merge_delta,
                        interval_left_to_new_interval,
                    )
                )
            elif interval_left_to_new_interval is None:
                criterion_delta_for_one_adjacent_interval_merge(
                    intervals, interval_to_be_merged, index_passed=False
                )
                best_merges.add(
                    (
                        interval_to_be_merged.right_merge_delta,
                        interval_to_be_merged,
                    )
                )
            else:
                best_merges.remove(
                    (
                        interval_left_to_new_interval.right_merge_delta,
                        interval_left_to_new_interval,
                    )
                )
                criterion_delta_for_one_adjacent_interval_merge(
                    intervals,
                    interval_left_to_new_interval,
                    index_passed=False,
                )
                best_merges.add(
                    (
                        interval_left_to_new_interval.right_merge_delta,
                        interval_left_to_new_interval,
                    )
                )
                criterion_delta_for_one_adjacent_interval_merge(
                    intervals, interval_to_be_merged, index_passed=False
                )
                best_merges.add(
                    (
                        interval_to_be_merged.right_merge_delta,
                        interval_to_be_merged,
                    )
                )
        else:
            break
    stop_counter(3)


def merge(interval, intervals, number_of_merges=1):
    neighbours_to_merge = [interval]
    merged_intervals = interval.nitj.copy()
    sum_of_old_priors_and_likelihoods = interval.sum_of_priors_and_likelihoods
    for i in range(number_of_merges):
        last_interval = neighbours_to_merge[-1].next
        neighbours_to_merge.append(last_interval)
        merged_intervals = list(
            map(add, neighbours_to_merge[-1].nitj.copy(), merged_intervals)
        )
        sum_of_old_priors_and_likelihoods += neighbours_to_merge[
            -1
        ].sum_of_priors_and_likelihoods

    interval.nitj = merged_intervals
    interval.included_right_frontier = neighbours_to_merge[
        -1
    ].included_right_frontier
    # NOW WE HAVE TO SEARCH for the old values of the sum of prior and likelihoods !!!!
    left_interval_node_criterion = interval.calculate_priors_and_likelihoods(
        mode="MergeAndUpdate"
    )  # it will update w, Priors, lkelihoods and sum_of_priors_and_likelihoods
    intervals.modl_value = (
        intervals.modl_value
        - sum_of_old_priors_and_likelihoods
        - log_binomial_coefficient(
            intervals.n + intervals.i - 1, intervals.i - 1
        )
        - ((intervals.i) * log(2))
        + log_binomial_coefficient(
            intervals.n + intervals.i - number_of_merges - 1,
            intervals.i - number_of_merges - 1,
        )
        + ((intervals.i - number_of_merges) * log(2))
        + left_interval_node_criterion
    )

    for i in range(
        1, len(neighbours_to_merge)
    ):  # Note the first element is the current interval that we are merging, no need to remove it!
        intervals.remove_interval(neighbours_to_merge[i])


def split_interval(
    interval, intervals, data, i
):  # i is interval index in intervalsList
    if interval == intervals.head:
        including_left_border = True
        left_bound = data[0][0]
    else:
        including_left_border = False
        left_bound = interval.previous.included_right_frontier
    right_bound = interval.included_right_frontier
    unique_values_in_both_intervals = list(
        data.irange_key(left_bound, right_bound, (including_left_border, True))
    )
    unique_values_in_both_intervals = list(
        map(itemgetter(0), unique_values_in_both_intervals)
    )
    unique_values_in_both_intervals = list(
        set(unique_values_in_both_intervals)
    )
    unique_values_in_both_intervals.sort()
    splits = {}
    left_and_right_interval_of_splits = {}
    previous_left_interval = [0, 0, 0, 0]
    prev_val = None

    for val in unique_values_in_both_intervals:
        if prev_val is None:
            left_split = list(
                data.irange_key(left_bound, val, (including_left_border, True))
            )
            left_interval = [0, 0, 0, 0]
            for interval_list in left_split:
                left_interval[
                    int((interval_list[1] * 2) + interval_list[2])
                ] += 1
        else:
            left_split = list(data.irange_key(prev_val, val, (False, True)))
            left_interval = [0, 0, 0, 0]
            for interval_list in left_split:
                left_interval[
                    int((interval_list[1] * 2) + interval_list[2])
                ] += 1
            left_interval = list(
                map(add, previous_left_interval, left_interval)
            )

        prev_val = val
        previous_left_interval = left_interval

        right_interval = list(map(sub, interval.nitj, left_interval))

        Left_interval = _Interval([left_interval, val, 0])
        Right_interval = _Interval(
            [right_interval, interval.included_right_frontier, 0]
        )

        criterion_one = Left_interval.calculate_priors_and_likelihoods(
            mode="MergeAndUpdate"
        )
        criterion_two = Right_interval.calculate_priors_and_likelihoods(
            mode="MergeAndUpdate"
        )

        split_criterion_val_left_and_right = (
            intervals.modl_value
            - interval.sum_of_priors_and_likelihoods
            - log_binomial_coefficient(
                intervals.n + intervals.i - 1, intervals.i - 1
            )
            - ((intervals.i) * log(2))
            + criterion_one
            + criterion_two
            + log_binomial_coefficient(intervals.n + intervals.i, intervals.i)
            + ((intervals.i + 1) * log(2))
        )

        if split_criterion_val_left_and_right < intervals.modl_value:
            splits[val] = split_criterion_val_left_and_right
            left_and_right_interval_of_splits[val] = [
                left_interval,
                right_interval,
            ]
    split_done = False
    best_split = None
    if splits:
        best_split = min(splits, key=splits.get)  # To be optimized maybe
        left_interval = left_and_right_interval_of_splits[best_split][0]
        right_interval = left_and_right_interval_of_splits[best_split][1]

        Left_interval = interval
        right_bound_of_the_right_interval = interval.included_right_frontier
        intervals.insert(
            _Interval([right_interval, right_bound_of_the_right_interval, 0]),
            i + 1,
        )
        Right_interval = intervals.get_nth(i + 1)

        Left_interval.nitj = left_interval
        Left_interval.included_right_frontier = best_split

        Left_interval.calculate_priors_and_likelihoods(
            mode="MergeAndUpdate"
        )  # it will update w, Priors, lkelihoods and sum_of_priors_and_likelihoods
        Right_interval.calculate_priors_and_likelihoods(
            mode="MergeAndUpdate"
        )  # it will update w, Priors, lkelihoods and sum_of_priors_and_likelihoods

        intervals.modl_value = splits[best_split]
        split_done = True
    return split_done, best_split, intervals


def post_optimization_to_be_repeated(intervals, data, i=0):
    data = SortedKeyList(data, key=itemgetter(0))
    interval = intervals.head
    old_modl_criterion_value = intervals.modl_value
    criterion_value_after_optimization = None
    looping_count = 0
    # Merge Split
    while True:
        looping_count += 1
        i = 0
        while True:
            interval = intervals.get_nth(i)
            if interval == 0:
                break
            split_done, split_val, intervals = split_interval(
                interval, intervals, data, i
            )
            if split_done:
                i += 2
            else:
                i += 1

        i = 0
        while True:
            interval = intervals.get_nth(i)
            if interval == 0:
                break
            if interval == intervals.tail:
                break
            if interval.next is None:
                # Arrived to the most left interval
                break

            old_split_val = interval.included_right_frontier
            merge(interval, intervals, 1)
            # Merge finished successfully
            split_after_merge_done, split_val, intervals = split_interval(
                interval, intervals, data, i
            )

            if split_after_merge_done and split_val != old_split_val:
                i += 2
            else:
                i += 1

        # MergeMegeSplit
        i = 0
        while True:
            interval = intervals.get_nth(i)

            if interval == 0:
                break
            if interval == intervals.tail:
                break
            if interval.next is None:
                # Arrived to the most left interval
                break
            if interval.next.next is None:
                break

            original_intervals_list = copy_list(intervals)
            merge(interval, intervals, 2)
            original_intervals_list_after_merge = copy_list(intervals)

            # Merge finished successfully
            split_after_merge_done, split_val, intervals = split_interval(
                interval, intervals, data, i
            )

            if (
                intervals.modl_value < original_intervals_list.modl_value
                and intervals.modl_value
                < original_intervals_list_after_merge.modl_value
            ):
                i += 2
            elif (
                original_intervals_list.modl_value
                < original_intervals_list_after_merge.modl_value
            ):
                intervals = copy_list(original_intervals_list)
                i += 3
            else:
                intervals = copy_list(original_intervals_list_after_merge)
                i += 1

        criterion_value_after_optimization = intervals.modl_value

        if round(criterion_value_after_optimization, 5) < round(
            old_modl_criterion_value, 5
        ):
            old_modl_criterion_value = criterion_value_after_optimization
            continue
        else:
            break
    return (
        intervals.size(),
        intervals.modl_value,
        intervals,
        intervals.get_discretization_info(),
    )


def copy_list(dll_to_be_copied):
    new_list = _DLL()  # head of the new list
    new_list.n = dll_to_be_copied.n
    new_list.modl_value = dll_to_be_copied.modl_value
    current = dll_to_be_copied.head  # used to iterate over the original list

    while current:
        new_list.append(current.get_interval_data())
        current = current.next

    return new_list


def calculate_feature_level(intervals, method="ED",get_intervals_uplift=True, min_val_in_var='-inf',max_val_in_var='+inf'):
    interval = intervals.head
    absolute_sum = 0
    
    intervals_vs_uplift=[] #a list of lists, each internal list contains the right boundary and the uplift
    
    if intervals.i == 1:
        if get_intervals_uplift:
            intervals_vs_uplift.append(["] "+str(min_val_in_var)+", " +str(max_val_in_var)+" ]",0,0,0])
            return 0,intervals_vs_uplift
        return 0
    
    previous_frontier=str(round(min_val_in_var,4))
    while interval:
        nitj = interval.nitj
        nit0j0 = nitj[0]
        nit0j1 = nitj[1]
        nit1j0 = nitj[2]
        nit1j1 = nitj[3]
        Ni = nit0j0 + nit0j1 + nit1j0 + nit1j1

        try:
            piYT1 = nit1j1 / (nit1j1 + nit1j0)
        except Exception:
            piYT1 = 0
        try:
            piYT0 = nit0j1 / (nit0j1 + nit0j0)
        except Exception:
            piYT0 = 0
        
        if get_intervals_uplift:
            interval_boundaries = "] "+previous_frontier+" , "+str(round(interval.included_right_frontier,4))+" ]"
            intervals_vs_uplift.append([interval_boundaries,round(piYT1,4),round(piYT0,4),round(round(piYT1,4)-round(piYT0,4),4)])
            previous_frontier=str(round(interval.included_right_frontier,4))
            
        if method == "ED":
            absolute_sum += (((piYT1) - (piYT0)) ** 2) * Ni / intervals.n  # ED
        elif method == "Chi":
            if piYT0 < 0.1**6:
                piYT0 = 0.1**6
            absolute_sum += (
                ((((piYT1) - (piYT0)) ** 2) / piYT0) * Ni / intervals.n
            )
        elif method == "KL":
            if piYT0 < 0.1**6:
                piYT0 = 0.1**6
            elif piYT0 > 1 - 0.1**6:
                piYT0 = 1 - 0.1**6
            absolute_sum += ((piYT1) * log(piYT1 / piYT0)) * Ni / intervals.n
        interval = interval.next
    
    if get_intervals_uplift:
        return absolute_sum,intervals_vs_uplift
    else:
        return absolute_sum


def execute_greedy_search_and_post_opt(df,get_intervals_uplift=True):
    col_name=df.columns[0]
    min_val_in_col=df[col_name].min() # A not very clean approach to add the boundaries in the final details dataframe using the calculate_feature_level()
    max_val_in_var=df[col_name].max()
    treatment_col_name = df.columns[1]
    y_name = df.columns[2]

    df = df.astype({treatment_col_name:'int'})
    df = df.astype({y_name:'int'})

    df = df.values.tolist()
    df = sorted(df, key=itemgetter(0))
    intervals = _DLL()
    intervals.n = len(df)

    intervals, intervals.modl_value, w = create_elementary_discretization(
        intervals, df
    )  # Elementary discretization

    compute_criterion_delta_for_all_possible_merges(
        intervals
    )  # Compute the cost of all possible merges of two adjacent intervals

    best_merges = intervals.get_sorted_list_of_address_and_right_merge_value(
        intervals.head
    )  # Get all the costs of 'all possible merges of two adjacent intervals' sorted
    best_merges = SortedKeyList(best_merges, key=itemgetter(0))

    # Start greedy search
    greedy_search(best_merges, intervals, intervals.n)

    # Post Optimization steps
    (
        intervals_num,
        umodl_val,
        intervals,
        info,
    ) = post_optimization_to_be_repeated(intervals, df)

    bounds = info[2]
    
    if get_intervals_uplift:
        feature_level_ed, intervals_vs_uplift = calculate_feature_level(intervals,get_intervals_uplift=get_intervals_uplift, min_val_in_var=round(min_val_in_col,4),max_val_in_var=round(max_val_in_var,4))
        return [feature_level_ed, bounds,col_name, intervals_vs_uplift]
    else:
        feature_level_ed = calculate_feature_level(intervals,get_intervals_uplift=False)
        return [feature_level_ed, bounds,col_name]

    



######################################################################################
# Copyright (c) 2023 Orange - All Rights Reserved                             #
# * This software is the confidential and proprietary information of Orange.         #
# * You shall not disclose such Restricted Information and shall use it only in      #
#   accordance with the terms of the license agreement you entered into with Orange  #
#   named the "kuplift - Python Library Evaluation License".                          #
# * Unauthorized copying of this file, via any medium is strictly prohibited.        #
# * See the "LICENSE.md" file for more details.                                      #
######################################################################################
import bisect
from math import log
import pandas as pd
import time
from .helperfunctions import (
    _log_fact_table,
    log_fact,
    log_binomial_coefficient,
)
from operator import itemgetter, add, sub
from sortedcontainers import SortedKeyList

_nb_counter = []
_start_counter = []
_start_time_counter = []
_deltatime_counter = []
_number_of_counters = 25
for i in range(_number_of_counters):
    _nb_counter.append(0)
    _start_counter.append(False)
    _start_time_counter.append(time.time())
    _deltatime_counter.append(0)


def start_counter(i):
    _nb_counter[i] = _nb_counter[i] + 1
    _start_counter[i] = True
    _start_time_counter[i] = time.time()


def stop_counter(i):
    _start_counter[i] = False
    diff = time.time() - _start_time_counter[i]
    _deltatime_counter[i] = _deltatime_counter[i] + diff
    _start_time_counter[i] = time.time()


def calc_criterion(nitj_interval, null_model=False):
    nitj_interval_sum = sum(nitj_interval)
    Fact_Class0Freq = _log_fact_table[(nitj_interval[0] + nitj_interval[2])]
    Fact_Class1Freq = _log_fact_table[(nitj_interval[1] + nitj_interval[3])]

    Fact_T0Freq = _log_fact_table[(nitj_interval[0] + nitj_interval[1])]
    Fact_T1Freq = _log_fact_table[(nitj_interval[2] + nitj_interval[3])]

    Fact_T0Class0Freq = _log_fact_table[nitj_interval[0]]
    Fact_T0Class1Freq = _log_fact_table[nitj_interval[1]]

    Fact_T1Class0Freq = _log_fact_table[nitj_interval[2]]
    Fact_T1Class1Freq = _log_fact_table[nitj_interval[3]]

    # Likelihood 1 w=0
    start_counter(0)
    likelihood_denum = 0
    j = 0
    likelihood_denum += Fact_Class0Freq
    j = 1
    likelihood_denum += Fact_Class1Freq

    likelihood_one_tmp = _log_fact_table[nitj_interval_sum] - likelihood_denum

    # Likelihood 2 w=1
    res_t = 0
    t = 0
    likelihood_denum = 0

    j = 0
    likelihood_denum += Fact_T0Class0Freq
    j = 1
    likelihood_denum += Fact_T0Class1Freq

    res_t += Fact_T0Freq - likelihood_denum
    t = 1
    likelihood_denum = 0

    j = 0
    likelihood_denum += Fact_T1Class0Freq
    j = 1
    likelihood_denum += Fact_T1Class1Freq

    res_t += Fact_T1Freq - likelihood_denum

    likelihood_two_tmp = res_t

    # Prior 1 w=0
    prior_one_tmp = log_binomial_coefficient(nitj_interval_sum + 1, 1)
    # Prior 2 w=1
    res_t = 0
    t = 0
    res_t_temp = log_binomial_coefficient(
        (nitj_interval[2 * t] + nitj_interval[2 * t + 1]) + 1, 1
    )
    res_t += res_t_temp
    t = 1
    res_t_temp = log_binomial_coefficient(
        (nitj_interval[2 * t] + nitj_interval[2 * t + 1]) + 1, 1
    )
    res_t += res_t_temp

    prior_two_tmp = res_t
    stop_counter(0)

    right_merge_w = None
    if not null_model:
        if (prior_one_tmp + likelihood_one_tmp) < (
            prior_two_tmp + likelihood_two_tmp
        ):
            right_merge_w = 0
        else:
            right_merge_w = 1
    else:
        if (prior_one_tmp + likelihood_one_tmp) > (
            prior_two_tmp + likelihood_two_tmp
        ):
            right_merge_w = 0
        else:
            right_merge_w = 1

    prior_one = (1 - right_merge_w) * prior_one_tmp
    prior_two = right_merge_w * prior_two_tmp
    likelihood_one = (1 - right_merge_w) * likelihood_one_tmp
    likelihood_two = (right_merge_w) * likelihood_two_tmp
    sum_of_priors_and_likelihoods = (
        prior_one + prior_two + likelihood_one + likelihood_two
    )
    return sum_of_priors_and_likelihoods


def split_interval(
    df,
    col_name,
    treatment_col_name,
    output_col_name,
    null_model_value,
    granularite=16,
):  # i is interval index in IntervalsList
    data = df[[col_name, treatment_col_name, output_col_name]].values.tolist()

    data = SortedKeyList(data, key=itemgetter(0))
    data_nitj = [0, 0, 0, 0]  # The frequency of treatment class
    for interval_list in data:
        data_nitj[int((interval_list[1] * 2) + interval_list[2])] += 1

    N = len(data)
    including_left_border = True
    left_bound = data[0][0]  # The smallest value
    right_bound = data[-1][0]  # The biggest value

    # Get all the unique values in the data i.e All unique values between left
    # and right bounds
    unique_values_in_both_intervals = list(
        data.irange_key(left_bound, right_bound, (including_left_border, True))
    )
    unique_values_in_both_intervals = list(
        map(itemgetter(0), unique_values_in_both_intervals)
    )
    unique_values_in_both_intervals = list(
        set(unique_values_in_both_intervals)
    )
    unique_values_in_both_intervals.sort()  # Sort the unique values

    splits = {}

    previous_left_interval = [0, 0, 0, 0]
    prev_val = None

    # Classical prior vs new prior of rissannen !!!
    prior_rissanen = log(2) + log_binomial_coefficient(N + 1, 1) + 2 * log(2)

    for val in unique_values_in_both_intervals:
        if (len(unique_values_in_both_intervals) <= 1) or (val == right_bound):
            break

        if prev_val is None:  # Enters here only for the first unique value
            # Get a list of data between left_bound and current unique value
            left_split = list(data.irange_key(left_bound, val, (True, True)))
            left_interval = [0, 0, 0, 0]
            for interval_list in left_split:
                left_interval[
                    int((interval_list[1] * 2) + interval_list[2])
                ] += 1
        else:
            left_split = list(data.irange_key(prev_val, val, (False, True)))
            left_interval = [0, 0, 0, 0]
            for interval_list in left_split:
                left_interval[
                    int((interval_list[1] * 2) + interval_list[2])
                ] += 1
            """
            New Left Interval frequencies is the sum of the previous left
            interval (bounded between Smallest value and prev_val) and the
            new left interval (bounded between prev_val and val)
            """
            left_interval = list(
                map(add, previous_left_interval, left_interval)
            )

        # the nitj for the right split (Which we call the right_interval)
        # will be the difference between the old nitj and the left_interval
        prev_val = val
        previous_left_interval = left_interval.copy()

        """
        The new rigt interval is the soustraction of all the data
        and the new left interval
        """
        right_interval = list(map(sub, data_nitj, left_interval))

        # Calculate criterion manually
        start_counter(22)
        criterion_one = calc_criterion(left_interval)  # prior and likelihood
        criterion_two = calc_criterion(right_interval)  # prior and likelihood
        stop_counter(22)

        # MODL value
        start_counter(24)
        split_criterion_val_left_and_right = (
            prior_rissanen + criterion_one + criterion_two
        )
        stop_counter(24)
        # If the MODL value is smaller than the null model value add it
        # to the splits dictionary
        if split_criterion_val_left_and_right < null_model_value:
            splits[val] = split_criterion_val_left_and_right
    best_split = None

    # If dictionary splits contain value, get the minimal one
    if splits:
        best_split = min(splits, key=splits.get)  # To be optimized maybe
        left_split = list(
            data.irange_key(left_bound, best_split, (True, True))
        )
        left_interval = [0, 0, 0, 0]
        for interval_list in left_split:
            left_interval[int((interval_list[1] * 2) + interval_list[2])] += 1
        right_interval = list(map(sub, data_nitj, left_interval))

        index_of_last_row_in_left_data = bisect.bisect_right(
            df[col_name].tolist(), best_split
        )
        left_data = df.iloc[:index_of_last_row_in_left_data, :]
        right_data = df.iloc[index_of_last_row_in_left_data:, :]
        return left_data, right_data, best_split, splits[best_split]
    else:
        split_criterion_val_left_and_right = None
    return -1


def calc_null_model(dff, att, treatment_col_name, output_col_name):
    data = dff[[att, treatment_col_name, output_col_name]].values.tolist()
    data = SortedKeyList(data, key=itemgetter(0))
    data_nitj = [0, 0, 0, 0]  # The frequency of treatment class
    for interval_list in data:
        data_nitj[int((interval_list[1] * 2) + interval_list[2])] += 1
    return (2 * log(2)) + calc_criterion(data_nitj)


def exec(df, attributeToDiscretize, treatment_col_name, output_col_name):
    null_model_value = calc_null_model(
        df, attributeToDiscretize, treatment_col_name, output_col_name
    )
    return split_interval(
        df,
        attributeToDiscretize,
        treatment_col_name,
        output_col_name,
        null_model_value,
    )


def umodl_binary_discretization(data, T, Y, attributeToDiscretize):
    df = pd.DataFrame()
    df = data.copy()
    treatment_col_name = T.name
    output_col_name = Y.name
    df[treatment_col_name] = T
    df[output_col_name] = Y
    df.sort_values(by=attributeToDiscretize, inplace=True)
    df.reset_index(inplace=True, drop=True)
    log_fact(df.shape[0] + 1)
    return exec(df, attributeToDiscretize, treatment_col_name, output_col_name)
